"""
A class hierarchy that measures moments of the observed emission lines.

----

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""
import inspect
from pathlib import Path
import logging

import numpy

from astropy.io import fits
import astropy.constants

from ..util.fitsutil import DAPFitsUtil
from ..util.bitmask import BitMask
from ..util.dapbitmask import DAPBitMask
from ..util.datatable import DataTable
from ..util.pixelmask import SpectralPixelMask
from ..util.log import log_output
from ..par.parset import KeywordParSet
from ..par.artifactdb import ArtifactDB
from ..par.emissionmomentsdb import EmissionMomentsDB
from .spatiallybinnedspectra import SpatiallyBinnedSpectra
from .stellarcontinuummodel import StellarContinuumModel
from .bandpassfilter import passband_integral, passband_integrated_width
from .bandpassfilter import passband_weighted_mean, passband_weighted_sdev, pseudocontinuum
from .bandpassfilter import emission_line_equivalent_width
from .spectralfitting import EmissionLineFit


class EmissionLineMomentsDef(KeywordParSet):
    """
    A class that holds the parameters necessary to perform the
    emission-line moment measurements.

    The defined parameters are:

    .. include:: ../tables/emissionlinemomentsdef.rst
    """
    def __init__(self, key='EMOMMPL11', minimum_snr=0.0, pixelmask=None, passbands='ELBMPL9',
                 redo_postmodeling=True, fit_vel_name='Ha-6564', overwrite=False):

        # Use the signature to get the parameters and the default values
        sig = inspect.signature(self.__class__)
        pars = list(sig.parameters.keys())
        defaults = [sig.parameters[key].default for key in pars]

        # Remaining definitions done by hand
        in_fl = [int, float]

        values = [key, minimum_snr, pixelmask, passbands, redo_postmodeling, fit_vel_name,
                  overwrite]
        dtypes = [str, in_fl, SpectralPixelMask, str, bool, str, bool]
        descr = ['Keyword used to distinguish between different emission-line moment databases.',
                 'Minimum S/N of spectrum to analyze',
                 'Object used to mask spectral pixels',
                 'Either a string identifying the emission-line bandpass filter database to ' \
                    'use, or the direct path to the parameter file defining the database',
                 'Redo the moment measurements after the emission-line modeling has been ' \
                    'performed',
                 'The name of the emission line used to set the redshift of each spaxel used ' \
                    'to set the observed wavelength of the bandpasses.',
                 'If the output file already exists, redo all the calculations and overwrite it.']
        super().__init__(pars, values=values, defaults=defaults, dtypes=dtypes, descr=descr)
        self.momdb = None
        self._validate()

    @classmethod
    def from_dict(cls, d):
        """
        Instantiate from a dictionary.
        """
        # Copy over the primary keywords.
        # TODO: Get these from inspection of the instatiation signature!
        _d = {}
        for key in ['key', 'minimum_snr', 'passbands', 'redo_postmodeling', 'fit_vel_name',
                    'overwrite']:
            if key in d.keys():
                _d[key] = d[key]

        artifacts = None if 'artifact_mask' not in d.keys() or d['artifact_mask'] is None \
                        else ArtifactDB.from_key(d['artifact_mask'])
        waverange = None if 'waverange' not in d.keys() or d['waverange'] is None \
                        else d['waverange']

        if not all([a is None for a in [artifacts, waverange]]):
            _d['pixelmask'] = SpectralPixelMask(artdb=artifacts, waverange=waverange)

        # Return the instantiation
        return super().from_dict(_d)

    def _validate(self):
        """
        Validate the object and instantiate the passband database to use for the
        emission-line moments.
        """
        if self['passbands'] is None:
            raise ValueError('Must define a passband database.')
        # Check if the provided string is a file
        parfile = Path(self['passbands']).resolve()
        if parfile.exists():
            self.momdb = EmissionMomentsDB(str(parfile))
            return

        # Otherwise, assume this is a keyword selecting a database distributed
        # with the DAP has been chosen.
        self.momdb = EmissionMomentsDB.from_key(self['passbands'])


class EmissionLineMomentsBitMask(DAPBitMask):
    r"""
    Derived class that specifies the mask bits for the emission-line
    moment measurements.  The maskbits defined are:
    
    .. include:: ../tables/emissionlinemomentsbitmask.rst
    """
    cfg_root = 'emission_line_moments_bits'


class EmissionLineMomentsDataTable(DataTable):
    """
    Primary data table with the results of the emission-line-moment
    calculations.

    Table includes:

    .. include:: ../tables/emissionlinemomentsdatatable.rst

    Args:
        neml (:obj:`int`):
            Number of emission lines being measured.
        bitmask (:class:`~mangadap.util.bitmask.BitMask`, optional):
            Object used to flag mask bits. If None, flags are simply
            boolean.
        shape (:obj:`int`, :obj:`tuple`, optional):
            The shape of the initial array. If None, the data array
            will not be instantiated; use :func:`init` to initialize
            the data array after instantiation.
    """
    def __init__(self, neml=1, bitmask=None, shape=None):
        # NOTE: This should require python 3.7 to make sure that this
        # is an "ordered" dictionary.
        datamodel = dict(BINID=dict(typ=int, shape=None, descr='Spectrum/Bin ID number'),
                         BINID_INDEX=dict(typ=int, shape=None,
                                          descr='Index of the spectrum in the list of '
                                                'provided spectra.'),
                         REDSHIFT=dict(typ=float, shape=None,
                                       descr='Redshift used for shifting the passbands'),
                         MASK=dict(typ=bool if bitmask is None else bitmask.minimum_dtype(),
                                   shape=(neml,),
                                   descr='Bad-value boolean or bit mask value for the moments'),
                         BCEN=dict(typ=float, shape=(neml,), descr='Center of the blue sideband.'),
                         BCONT=dict(typ=float, shape=(neml,),
                                    descr='Pseudo-continuum in the blue sideband'),
                         BCONTERR=dict(typ=float, shape=(neml,),
                                       descr='Error in the blue-sideband pseudo-continuum'),
                         RCEN=dict(typ=float, shape=(neml,), descr='Center of the red sideband.'),
                         RCONT=dict(typ=float, shape=(neml,),
                                    descr='Pseudo-continuum in the red sideband'),
                         RCONTERR=dict(typ=float, shape=(neml,),
                                       descr='Error in the red-sideband pseudo-continuum'),
                         CNTSLOPE=dict(typ=float, shape=(neml,),
                                       descr='Continuum slope used to determine the continuum at '
                                             'the line center.'),
                         FLUX=dict(typ=float, shape=(neml,), descr='Summed flux (0th moment)'),
                         FLUXERR=dict(typ=float, shape=(neml,),
                                      descr='Error in the summed flux'),
                         MOM1=dict(typ=float, shape=(neml,),
                                   descr='Line centroid redshift (:math:`cz`; 1st moment)'),
                         MOM1ERR=dict(typ=float, shape=(neml,),
                                      descr='Error in the line centroid redshift'),
                         MOM2=dict(typ=float, shape=(neml,),
                                   descr='Line standard deviation (2nd moment)'),
                         MOM2ERR=dict(typ=float, shape=(neml,),
                                      descr='Error in the line standard deviation'),
                         SINST=dict(typ=float, shape=(neml,),
                                    descr='Instrumental dispersion at the line centroid'),
                         BMED=dict(typ=float, shape=(neml,),
                                   descr='Median flux in the blue sideband used for EW'),
                         RMED=dict(typ=float, shape=(neml,),
                                   descr='Median flux in the red sideband used for EW'),
                         EWCONT=dict(typ=float, shape=(neml,),
                                     descr='Continuum value used for EW calculation'),
                         EW=dict(typ=float, shape=(neml,),
                                 descr='Equivalent width (FLUX/pseudo continuum)'),
                         EWERR=dict(typ=float, shape=(neml,),
                                    descr='Error in the equivalent width'))

        keys = list(datamodel.keys())
        super(EmissionLineMomentsDataTable,
                self).__init__(keys, [datamodel[k]['typ'] for k in keys],
                               element_shapes=[datamodel[k]['shape'] for k in keys],
                               descr=[datamodel[k]['descr'] for k in keys],
                               shape=shape)


class EmissionLineMoments:
    r"""
    Class that holds the emission-line moment measurements.

    Args:
        loggers (list): (**Optional**) List of `logging.Logger`_ objects
            to log progress; ignored if quiet=True.  Logging is done
            using :func:`mangadap.util.log.log_output`.  Default is no
            logging.
        quiet (bool): (**Optional**) Suppress all terminal and logging
            output.  Default is False.

    Attributes:
        loggers (list): List of `logging.Logger`_ objects to log
            progress; ignored if quiet=True.  Logging is done using
            :func:`mangadap.util.log.log_output`.
        quiet (bool): Suppress all terminal and logging output.

    """
#    @profile
    def __init__(self, database, binned_spectra, stellar_continuum=None,
                 emission_line_model=None, redshift=None, database_list=None, artifact_path=None,
                 bandpass_path=None, output_path=None, output_file=None, hardcopy=True,
                 overwrite=None, checksum=False, loggers=None, quiet=False):

        self.loggers = None
        self.quiet = False

        # Define the database properties
#        self.database = None
#        self.pixelmask = None
#        self.artdb = None
#        self.momdb = None

        self.database = database
        if not isinstance(self.database, EmissionLineMomentsDef):
            raise TypeError('Method must have type EmissionLineMomentsDef.')

#        self._define_databases(database_key, database_list=database_list,
#                               artifact_path=artifact_path, bandpass_path=bandpass_path)

        self.nmom = self.database.momdb.size

        self.binned_spectra = None
        self.stellar_continuum = None
        self.emission_line_model = None
        self.redshift = None

        # Define the output directory and file
        self.directory_path = None      # Set in default_paths
        self.output_file = None
        self.hardcopy = None

        # Initialize the objects used in the assessments
        self.bitmask = EmissionLineMomentsBitMask()

        self.hdu = None
        self.checksum = checksum
        self.spatial_shape = None
        self.nspec = None
        self.spatial_index = None
        self.image_arrays = None
        self._assign_image_arrays()

        self.nbins = None
        self.missing_bins = None

        # Run the assessments of the DRP file
        self.measure(binned_spectra, stellar_continuum=stellar_continuum,
                     emission_line_model=emission_line_model, redshift=redshift, 
                     output_path=output_path, output_file=output_file, hardcopy=hardcopy,
                     overwrite=overwrite, loggers=loggers, quiet=quiet)

    def __getitem__(self, key):
        return self.hdu[key]

    @staticmethod
    def default_paths(cube, method_key, rdxqa_method, binning_method, stelcont_method=None,
                      elmodel_method=None, output_path=None, output_file=None):
        """
        Set the default directory and file name for the output file.

        Args:
            cube (:class:`mangadap.datacube.datacube.DataCube`):
                Datacube to analyze.
            method_key (:obj:`str`):
                Keyword designating the method used for the reduction
                assessments.
            rdxqa_method (:obj:`str`):
                The method key for the basic assessments of the datacube.
            binning_method (:obj:`str`):
                The method key for the spatial binning.
            stelcont_method (:obj:`str`, optional):
                The method key for the stellar-continuum fitting method.  If
                None, not included in output file name.
            elmodel_method (:obj:`str`, optional):
                The method key for the emission-line modeling method.  If None,
                not included in the output file name.
            output_path (:obj:`str`, `Path`_, optional):
                The path for the output file.  If None, the current working
                directory is used.
            output_file (:obj:`str`, optional):
                The name of the output "reference" file. The full path of the
                output file will be :attr:`directory_path`/:attr:`output_file`.
                If None, the default is to combine ``cube.output_root`` and the
                method keys.  The order of the keys is the order of operations
                (rdxqa, binning, stellar continuum, emission-line model).

        Returns:
            :obj:`tuple`: Returns a `Path`_ with the output directory and a
            :obj:`str` with the output file name.
        """
        directory_path = Path('.').resolve() if output_path is None \
                                else Path(output_path).resolve()
        method = f'{rdxqa_method}-{binning_method}'
        if stelcont_method is not None:
            method = f'{method}-{stelcont_method}'
        if elmodel_method is not None:
            method = f'{method}-{elmodel_method}'
        method = f'{method}-{method_key}'
        _output_file = f'{cube.output_root}-{method}.fits.gz' if output_file is None \
                            else output_file
        return directory_path, _output_file


    def _initialize_primary_header(self, hdr=None, measurements_binid=None):
        """
        Construct the primary header for the reference file.

        Args:
            hdr (`astropy.io.fits.Header`_, optional):
                Input base header for added keywords. If None, uses
                the :attr:`cube` header (if there is one) and then
                cleans the header using
                :func:`mangadap.util.fitsutil.DAPFitsUtil.clean_dap_primary_header`.
            measurements_binid (`numpy.ndarray`_, optional):
                Bin IDs for moment measurements. Only use is to check
                if this is None, and the boolean result is saved to
                the header to indicate if the emission-line moment
                measurements are disconnected from the
                stellar-continuum measurements.

        Returns:
            `astropy.io.fits.Header`_: Initialized header object.
        """
        # Copy the from the DRP and clean it
        if hdr is None:
            hdr = self.binned_spectra.cube.prihdr.copy()
            hdr = DAPFitsUtil.clean_dap_primary_header(hdr)
        
        hdr['AUTHOR'] = 'Kyle B. Westfall <westfall@ucolick.org>'
        hdr['ELMKEY'] = (self.database['key'], 'Emission-line moments database keyword')
        hdr['ELMMINSN'] = (self.database['minimum_snr'], 'Minimum S/N of spectrum to include')
#        hdr['ARTDB'] = (self.database['artifacts'], 'Artifact database keyword')
        hdr['MOMDB'] = (self.database['passbands'], 'Emission-line moments database keyword')
        if self.stellar_continuum is not None:
            hdr['SCKEY'] = (self.stellar_continuum.method['key'], 'Stellar-continuum model keyword')
        if self.emission_line_model is not None:
            hdr['ELFKEY'] = (self.emission_line_model.method['key'],
                                'Emission-line modeling method keyword')
        hdr['NBINS'] = (self.nbins, 'Number of unique spectra for measurements')
        hdr['ELMREBIN'] = (measurements_binid is not None, 'Bin IDs disconnected from SC binning')
        return hdr

    def _assign_image_arrays(self):
        """
        Set :attr:`image_arrays`, which contains the list of extensions
        in :attr:`hdu` that are on-sky image data.
        """
        self.image_arrays = [ 'BINID' ]

    def _get_missing_bins(self, unique_bins=None):
        if unique_bins is None:
            good_snr = self.binned_spectra.above_snr_limit(self.database['minimum_snr'])
            return numpy.sort(self.binned_spectra.missing_bins + 
                        self.binned_spectra['BINS'].data['BINID'][numpy.invert(good_snr)].tolist())
        return SpatiallyBinnedSpectra._get_missing_bins(unique_bins)


    def _assign_redshifts(self, redshift, measure_on_unbinned_spaxels, good_snr,
                          default_redshift=None):
        """
        Set the redshift to use for each spectrum for the emission-line
        moments.
        
        In terms of precedence, directly provided redshifts override
        those in any available EmissionLineModel.

        If :attr:`emission_line_model` and redshift are None, the default
        redshift is used (or 0.0 if this is also None).

        To get the emission-line stellar kinematics, the function calls
        :func:`mangadap.proc.emissionlinemodel.EmissionLineModel.matched_kinematics`.
        The method must have a fit_vel_name define for this use!

        In this function, the provided redshift must be a single value
        or None; therefore, the means of any vectors should be provided
        instead of the full vector.

        Args:
            redshift (:obj:`float`, `numpy.ndarray`_):
                Redshifts (:math:`z`) to use for each spectrum.
            measure_on_unbinned_spaxels (:obj:`bool`):
                Flag that method expects to measure moments on unbinned
                spaxels.
            good_snr (`numpy.ndarray`_):
                Boolean array setting which spectra have sufficient S/N
                for the measurements.

            default_redshift (:obj:`float`, optional):

                Only used if there are stellar kinematics available.
                Provides the default redshift to use for spectra without
                stellar measurements; see ``redshift`` in
                :func:`mangadap.proc.stellarcontinuummodel.StellarContinuumModel.matched_kinematics`.
                If None (default), the median of the unmasked stellar
                velocities will be used.

        """

        # Construct the binid matrix if measuring on unbinned spaxels
        if measure_on_unbinned_spaxels:
            binid = numpy.full(self.spatial_shape, -1, dtype=int)
            binid.ravel()[good_snr] = numpy.arange(self.nbins)
            missing = []
            nspec = self.binned_spectra.cube.nspec
        else:
            binid = self.binned_spectra['BINID'].data
            missing = self.binned_spectra.missing_bins
            nspec = self.binned_spectra.nbins

        #---------------------------------------------------------------
        # Get the redshifts measured for the selected emission line and
        # use them if no default value is provided
        if self.emission_line_model is not None and self.database['fit_vel_name'] is not None \
                and redshift is None:
            self.redshift, _ = self.emission_line_model.matched_kinematics(
                                        binid, self.database['fit_vel_name'],
                                        redshift=default_redshift, nearest=True, missing=missing)
            if measure_on_unbinned_spaxels:
                tmp = self.redshift.copy()
                self.redshift = numpy.zeros(nspec, dtype=float)
                self.redshift[good_snr] = tmp
            return

        #---------------------------------------------------------------
        # Use the default value(s)
        _redshift = numpy.atleast_1d(redshift)
        if len(_redshift) not in [ 1, nspec ]:
            raise ValueError('Provided redshift must be either a single value or match the '
                             'number of binned spectra or the number of unbinned spaxels.')
        self.redshift = numpy.full(nspec, redshift, dtype=float) \
                                if len(_redshift) == 1 else _redshift.copy()

    def _flag_good_spectra(self, measure_on_unbinned_spaxels):
        if measure_on_unbinned_spaxels:
            return self.binned_spectra.check_fgoodpix()
        return self.binned_spectra.above_snr_limit(self.database['minimum_snr'])

    @staticmethod
    def sideband_pseudocontinua(wave, spec, sidebands, spec_n=None, err=None, log=True):
        """
        Get the side-band integrals.

        err is a single vector that is the same for all spec!

        Returns:
            :obj:`float`, `numpy.ndarray`_: Return five arrays or
            floats: (1) The center of each passband, (2) the mean
            continuum level, (3) the propagated error in the
            continuum level (will be None if no errors are provided),
            (4) flag that part of the passband was masked, (5) flag
            that the passband was fully masked or empty.
        """

        if spec.ndim == 1:
            return pseudocontinuum(wave, spec, passband=sidebands, err=err, log=log)

        # Calculate the pseudo-continua in the sidebands
        nspec = spec.shape[0]
        nbands = sidebands.shape[0]

        # More than one spectrum entered, use spec_n to decide which
        # spectrum to use for the integrals.  If spec_n is None and
        # there is more than one spectrum (improper use of the code!),
        # allow the code to continue but issue a warning
        if spec_n is None:
            warnings.warn('Multiple spectra entered, but no designation for which to use!')
            _spec_n = numpy.zeros(nbands, dtype=int)
        else:
            _spec_n = spec_n
        
        # TODO: Change this to a warning?
        if numpy.amax(_spec_n) > nspec-1:
            raise ValueError('Selected spectrum not provided!')

        # Instantiate the arrays
        continuum = numpy.zeros(nbands, dtype=float)
        continuum_error = None if err is None else numpy.zeros(nbands, dtype=float)
        band_center = numpy.zeros(nbands, dtype=float)
        band_incomplete = numpy.zeros(nbands, dtype=bool)
        band_empty = numpy.zeros(nbands, dtype=bool)

        for i in range(nspec):
            indx = _spec_n==i
            if numpy.sum(indx) == 0:
                continue
            band_center[indx], continuum[indx], _continuum_error, band_incomplete[indx], \
                    band_empty[indx] = pseudocontinuum(wave, spec[i,:], passband=sidebands[indx,:],
                                                       err=err, log=log)
            if err is not None:
                continuum_error[indx] = _continuum_error
        return band_center, continuum, continuum_error, band_incomplete, band_empty

    @staticmethod
    def single_band_moments(wave, spec, passband, restwave, err=None):
        """
        Measure the moments for a single band.

        If spectrum errors are not provided, moment errors are returned
        as 0.0.

        Args:
            wave (`numpy.ndarray`_):
                Wavelengths in angstroms
            spec (`numpy.ndarray`_):
                Flux in flux density (per angstrom)
            restwave (float):
                The rest wavelength of the line.
            err (`numpy.ndarray`_):
                The 1-sigma errors in the flux density.

        Returns:
            flux
            fluxerr
            mom1
            mom1err
            mom2
            mom2err
            incomplete band
            empty band
            division by zero
            undefined 1st moment
            undefined 2nd moment

        """
        _passband = numpy.atleast_1d(passband)
        if _passband.ndim > 1:
            raise ValueError('Should only pass a single band to single_band_moments.')
        if len(_passband) != 2:
            raise ValueError('Band should be a list of 2 numbers.')
        # Get the fraction of the passband that is unmasked
        interval_frac = passband_integrated_width(wave, spec, passband=passband, log=True) \
                                / numpy.diff(passband)[0]
        incomplete = interval_frac < 1.0
        empty = numpy.invert(interval_frac > 0.0)
        if empty:
            return 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, incomplete, empty, False, True, True

        # Get the redshift vector
        cz = astropy.constants.c.to('km/s').value*(wave/restwave-1)

        # Get the integrated flux
        flux = passband_integral(wave, spec, passband=passband, log=True)
        fluxerr = 0.0 if err is None else \
                    passband_integral(wave, err, passband=passband, log=True, quad=True)

        # No flux in the passband, meaning division by zero will occur!
        if not numpy.absolute(flux) > 0.0:
            return flux, fluxerr, 0.0, 0.0, 0.0, 0.0, incomplete, empty, True, True, True

        # Get the first and second moments:
        mom1, mom1err = passband_weighted_mean(wave, spec, cz, yerr=err, passband=passband,
                                               log=True)
        mom2, mom2err = passband_weighted_sdev(wave, spec, cz, yerr=err, passband=passband,
                                               log=True)

        # If no errors provided, set errors to 0
        if err is None:
            mom1err = 0.0
            mom2err = 0.0

        # Set masked values to 0.0
        undefined_mom1 = False
        if isinstance(mom1, numpy.ma.core.MaskedConstant) and mom1.mask:
            mom1 = 0.0
            mom1err = 0.0
            undefined_mom1 = True
        undefined_mom2 = False
        if isinstance(mom2, numpy.ma.core.MaskedConstant) and mom2.mask:
            mom2 = 0.0
            mom2err = 0.0
            undefined_mom2 = True
        if isinstance(mom1err, numpy.ma.core.MaskedConstant) and mom1err.mask:
            mom1err = 0.0
        if isinstance(mom2err, numpy.ma.core.MaskedConstant) and mom2err.mask:
            mom2err = 0.0

        return flux, fluxerr, mom1, mom1err, mom2, mom2err, incomplete, empty, False, \
                    undefined_mom1, undefined_mom2

    @staticmethod
    def continuum_subtracted_moments(wave, spec, mainbands, restwave, bcen, bcont, rcen, rcont,
                                     spec_n=None, err=None, sres=None):
        """
        Calculate the continuum-subtracted moments
        
        err and sres are single vectors that are the same for all spec!

        Returns:
            cntm
            cntb
            flux
            fluxerr
            mom1
            mom1err
            mom2
            mom2err
            sinst
            incomplete
            empty
            divbyzero
            undefined_mom2

        """
        # Get the parameters for the linear continuum across the
        # primary passband
        nbands = mainbands.shape[0]
        flux = numpy.zeros(nbands, dtype=float)
        fluxerr = numpy.zeros(nbands, dtype=float)
        mom1 = numpy.zeros(nbands, dtype=float)
        mom1err = numpy.zeros(nbands, dtype=float)
        mom2 = numpy.zeros(nbands, dtype=float)
        mom2err = numpy.zeros(nbands, dtype=float)
        incomplete = numpy.zeros(nbands, dtype=bool)
        empty = numpy.zeros(nbands, dtype=bool)
        divbyzero = numpy.zeros(nbands, dtype=bool)
        undefined_mom1 = numpy.zeros(nbands, dtype=bool)
        undefined_mom2 = numpy.zeros(nbands, dtype=bool)

        if numpy.any(rcen-bcen == 0):
            print('band centers are the same!')
        cntm = (rcont - bcont) / (rcen - bcen)
        cntb = bcont - bcen*cntm

        # If more than one spectrum entered, use spec_n to decide which
        # spectrum to use for the integrals
        if len(spec.shape) > 1:
            # In this case, spec_n should always be provided, but allow
            # the code to continue with a warning
            if spec_n is None:
                warnings.warn('Multiple spectra entered, but no designation for which to use!')
                _spec_n = numpy.zeros(self.nmom, dtype=int)
            else:
                _spec_n = spec_n

        # Get the moments for all the main passbands
        for i,p in enumerate(mainbands):
            cont = cntb[i] + cntm[i]*wave
            _spec = spec - cont if len(spec.shape) == 1 else spec[_spec_n[i],:] - cont

            flux[i], fluxerr[i], mom1[i], mom1err[i], mom2[i], mom2err[i], incomplete[i], \
                        empty[i], divbyzero[i], undefined_mom1[i], undefined_mom2[i] \
                        = EmissionLineMoments.single_band_moments(wave, _spec, p, restwave[i],
                                                                  err=err)

        # Get the instrumental dispersion at the position of the line
        # center
        sinst = numpy.zeros(nbands, dtype=float) if sres is None \
                        else EmissionLineFit.instrumental_dispersion(wave, sres, restwave, mom1)

        return cntm, cntb, flux, fluxerr, mom1, mom1err, mom2, mom2err, sinst, incomplete, empty, \
                    divbyzero, undefined_mom1, undefined_mom2

    @staticmethod
    def measure_moments(momdb, wave, flux, ivar=None, mask=None, sres=None, continuum=None,
                        redshift=None, bitmask=None):
        """
        Measure the emission-line moments.

        If not input as masked arrays, flux is converted to one.
        """
        # Check the input moment database
        if not isinstance(momdb, EmissionMomentsDB):
            raise TypeError('Input database must have type EmissionMomentsDB.')

        # Check the bitmask if provided
        if bitmask is not None and not isinstance(bitmask, BitMask):
            raise TypeError('Input bitmask must have type BitMask.')

        # Prepare the spectra for the measurements
        _flux, _err, _sres, _continuum, _redshift, _ \
                    = EmissionLineFit.check_and_prep_input(wave, flux, ivar=ivar, mask=mask,
                                                           sres=sres, continuum=continuum,
                                                           redshift=redshift)

        # Subtract the continuum and determine where the continuum is
        # undefined.  If _continuum is None, this just returns _fluxnc =
        # _flux and no_continuum = None.
        _fluxnc, no_continuum = EmissionLineFit.subtract_continuum(_flux, _continuum)

        # Initialize the output data
        nspec = _flux.shape[0]
        nmom = momdb.size
        measurements = EmissionLineMomentsDataTable(neml=nmom, bitmask=bitmask, shape=nspec)

        # Save the redshift used
        measurements['REDSHIFT'] = _redshift

        # Mask dummy lines
        if numpy.any(momdb.dummy):
            measurements['MASK'][:,momdb.dummy] = True if bitmask is None else \
                    bitmask.turn_on(measurements['MASK'][:,momdb.dummy], 'UNDEFINED_BANDS')
        
        # No valid lines
        if numpy.all(momdb.dummy):
            return measurements

        # Common arrays used for each spectrum
        blue_fraction = numpy.zeros(nmom, dtype=float)
        red_fraction = numpy.zeros(nmom, dtype=float)

        incomplete = numpy.zeros(nmom, dtype=bool)
        empty = numpy.zeros(nmom, dtype=bool)
        divbyzero = numpy.zeros(nmom, dtype=bool)
        undefined_mom1 = numpy.zeros(nmom, dtype=bool)
        undefined_mom2 = numpy.zeros(nmom, dtype=bool)

        # Perform the measurements for each spectrum
        for i in range(nspec):
            print('Measuring emission-line moments in spectrum: {0}/{1}'.format(i+1,nspec),
                  end='\r')

            incomplete[:] = False
            empty[:] = False
            divbyzero[:] = False
            undefined_mom1[:] = False
            undefined_mom2[:] = False

            # Shift the sidebands to the appropriate redshift
            _bluebands = momdb['blueside'][numpy.invert(momdb.dummy)]*(1.0+_redshift[i])
            _redbands = momdb['redside'][numpy.invert(momdb.dummy)]*(1.0+_redshift[i])
            __err = None if _err is None else _err[i,:]
            __sres = None if _sres is None else _sres[i,:]

            # Check if the model was subtracted over the full range of
            # the sidebands, or if the model was subtracted in one
            # sideband but not the other
            if no_continuum is None:
                spec = _flux[i,:]
                spec_n = None

                # Flag moment calculations without a stellar-absorption
                # correction
                measurements['MASK'][i,:] = True if bitmask is None \
                                                else bitmask.turn_on(measurements['MASK'][i,:],
                                                                     'NO_ABSORPTION_CORRECTION')
            else:
                blue_fraction[numpy.invert(momdb.dummy)] \
                        = passband_integrated_width(wave, no_continuum[i,:], passband=_bluebands,
                                                log=True) / numpy.diff(_bluebands, axis=1).ravel()
                indx = numpy.logical_and(blue_fraction > 0.0, blue_fraction < 1.0)
                measurements['MASK'][i,indx] = True if bitmask is None \
                            else bitmask.turn_on(measurements['MASK'][i,indx], 'BLUE_JUMP')

                red_fraction[numpy.invert(momdb.dummy)] \
                        = passband_integrated_width(wave, no_continuum[i,:], passband=_redbands,
                                                log=True) / numpy.diff(_redbands, axis=1).ravel()
                indx = numpy.logical_and(red_fraction > 0.0, red_fraction < 1.0)
                measurements['MASK'][i,indx] = True if bitmask is None \
                            else bitmask.turn_on(measurements['MASK'][i,indx], 'RED_JUMP')
                
                indx = (numpy.invert(blue_fraction > 0) & (red_fraction > 0)) \
                            | (numpy.invert(red_fraction > 0) & (blue_fraction > 0))
                measurements['MASK'][i,indx] = True if bitmask is None \
                            else bitmask.turn_on(measurements['MASK'][i,indx],'JUMP_BTWN_SIDEBANDS')

                # Only use model-subtracted measurements if there are no
                # jumps in the bands or between the bands
                jumps = measurements['MASK'][i,:] if bitmask is None \
                            else bitmask.flagged(measurements['MASK'][i,:],
                                                 flag=['BLUE_JUMP', 'RED_JUMP',
                                                       'JUMP_BTWN_SIDEBANDS'])

                spec = numpy.ma.append(_flux[i,:].reshape(1,-1), _fluxnc[i,:].reshape(1,-1), axis=0)
                spec_n = numpy.zeros(nmom, dtype=int)
                spec_n[numpy.invert(jumps)] = 1

                # Flag moment calculations as having not been corrected
                # for stellar absorption
                measurements['MASK'][i,spec_n == 0] = True if bitmask is None \
                            else bitmask.turn_on(measurements['MASK'][i,spec_n == 0],
                                                 'NO_ABSORPTION_CORRECTION')

            # Get the blue pseudo continuum
            measurements['BCEN'][i,numpy.invert(momdb.dummy)], \
                measurements['BCONT'][i,numpy.invert(momdb.dummy)], conterr, \
                incomplete[numpy.invert(momdb.dummy)], empty[numpy.invert(momdb.dummy)] \
                    = EmissionLineMoments.sideband_pseudocontinua(wave, spec, _bluebands,
                                                                  spec_n=None if spec_n is None
                                                            else spec_n[numpy.invert(momdb.dummy)],
                                                                  err=__err)
            if __err is not None:
                measurements['BCONTERR'][i,numpy.invert(momdb.dummy)] = conterr
            if numpy.sum(incomplete) > 0:
                measurements['MASK'][i,incomplete] = True if bitmask is None \
                            else bitmask.turn_on(measurements['MASK'][i,incomplete], 'BLUE_INCOMP')
            if numpy.sum(empty) > 0:
                measurements['MASK'][i,empty] = True if bitmask is None \
                            else bitmask.turn_on(measurements['MASK'][i,empty], 'BLUE_EMPTY')

            # Get the red pseudo continuum
            measurements['RCEN'][i,numpy.invert(momdb.dummy)], \
                measurements['RCONT'][i,numpy.invert(momdb.dummy)], conterr, \
                incomplete[numpy.invert(momdb.dummy)], empty[numpy.invert(momdb.dummy)] \
                    = EmissionLineMoments.sideband_pseudocontinua(wave, spec, _redbands,
                                                                  spec_n=None if spec_n is None
                                                            else spec_n[numpy.invert(momdb.dummy)],
                                                                  err=__err)
            if __err is not None:
                measurements['RCONTERR'][i,numpy.invert(momdb.dummy)] = conterr
            if numpy.sum(incomplete) > 0:
                measurements['MASK'][i,incomplete] = True if bitmask is None \
                            else bitmask.turn_on(measurements['MASK'][i,incomplete], 'RED_INCOMP')
            if numpy.sum(empty) > 0:
                measurements['MASK'][i,empty] = True if bitmask is None \
                            else bitmask.turn_on(measurements['MASK'][i,empty], 'RED_EMPTY')

            # Get the main passband integral after subtracting the
            # continuum
            indx = numpy.invert(momdb.dummy) \
                        & numpy.invert(measurements['MASK'][i,:] if bitmask is None
                                        else bitmask.flagged(measurements['MASK'][i,:],
                                                             flag=['BLUE_EMPTY', 'RED_EMPTY',
                                                                   'BLUE_JUMP', 'RED_JUMP']))

            _mainbands = momdb['primary'][indx]*(1.0+_redshift[i])

            measurements['CNTSLOPE'][i,indx], cntb, measurements['FLUX'][i,indx], \
                    measurements['FLUXERR'][i,indx], measurements['MOM1'][i,indx], \
                    measurements['MOM1ERR'][i,indx], measurements['MOM2'][i,indx], \
                    measurements['MOM2ERR'][i,indx], measurements['SINST'][i,indx], \
                    incomplete[indx], empty[indx], divbyzero[indx], undefined_mom1[indx], \
                    undefined_mom2[indx] = \
                            EmissionLineMoments.continuum_subtracted_moments(wave, spec,
                                                               _mainbands, momdb['restwave'][indx],
                                                               measurements['BCEN'][i,indx],
                                                               measurements['BCONT'][i,indx],
                                                               measurements['RCEN'][i,indx],
                                                               measurements['RCONT'][i,indx],
                                                               spec_n=None if spec_n is None 
                                                                    else spec_n[indx], err=__err,
                                                               sres=__sres) 

            # Turn on any necessary bits
            measurements['MASK'][i,incomplete] = True if bitmask is None \
                        else bitmask.turn_on(measurements['MASK'][i,incomplete], 'MAIN_INCOMP')
            measurements['MASK'][i,empty] = True if bitmask is None \
                        else bitmask.turn_on(measurements['MASK'][i,empty], 'MAIN_EMPTY')
            measurements['MASK'][i,divbyzero] = True if bitmask is None \
                        else bitmask.turn_on(measurements['MASK'][i,divbyzero], 'DIVBYZERO')
            measurements['MASK'][i,undefined_mom1] = True if bitmask is None \
                        else bitmask.turn_on(measurements['MASK'][i,undefined_mom1],
                                             'UNDEFINED_MOM1')
            measurements['MASK'][i,undefined_mom2] = True if bitmask is None \
                        else bitmask.turn_on(measurements['MASK'][i,undefined_mom2],
                                             'UNDEFINED_MOM2')

        print('Measuring emission-line moments in spectrum: {0}/{0}'.format(nspec))
        return measurements
        
    def file_name(self):
        """Return the name of the output file."""
        return self.output_file

    def file_path(self):
        """Return the full path to the output file."""
        if self.directory_path is None or self.output_file is None:
            return None
        return self.directory_path / self.output_file

    def measure(self, binned_spectra, stellar_continuum=None, emission_line_model=None,
                redshift=None, output_path=None, output_file=None, hardcopy=True, overwrite=None,
                loggers=None, quiet=False):
        """
        Measure the emission-line moments using the binned spectra.

        If neither stellar-continuum nor emission-line models are
        provided:

            - Moments are measure on the binned spectra
            - No continuum subtraction is performed

        If a stellar-continuum model is provided without an
        emission-line model:

            - Moments are measured on the binned spectra
            - The best-fitting stellar continuum is subtracted

        If an emission-line model is provided without a
        stellar-continuum model:

            - Moments are measured on the relevant (binned or unbinned)
              spectra
            - If the emission-line model includes data regarding the
              stellar-continuum fit (template spectra and template
              weights), the continuum is subtracted before the
              measurements are made; otherwise, no continuum is
              subtracted

        If both stellar-continuum and emission-line models are provided,
        and if the stellar-continuum and emission-line fits are
        performed on the same spectra:

            - Moments are measured on the relevant (binned or unbinned)
              spectra
            - If the emission-line model includes data regarding the
              stellar-continuum fit (template spectra and template
              weights), the continuum is subtracted before the
              measurements are made; otherwise, the stellar continuum is
              used.

        If both stellar-continuum and emission-line models are provided,
        and if the stellar-continuum and emission-line fits are
        performed on different spectra:

            - The behavior is exactly as if the stellar-continuum model
              was not provided.

        Args:
            output_path (:obj:`str`, `Path`_, optional):
                The path for the output file.  If None, the current working
                directory is used.
            output_file (:obj:`str`, optional):
                The name of the output "reference" file. The full path of the
                output file will be :attr:`directory_path`/:attr:`output_file`.
                If None, the default is to combine ``cube.output_root`` and the
                method keys.  The order of the keys is the order of operations
                (rdxqa, binning, stellar continuum, emission-line model).

        """
        # Initialize the reporting
        if loggers is not None:
            self.loggers = loggers
        self.quiet = quiet
        
        # SpatiallyBinnedSpectra object always needed
        if binned_spectra is None:
            raise ValueError('Must provide spectra object for fitting.')
        if not isinstance(binned_spectra, SpatiallyBinnedSpectra):
            raise TypeError('Must provide a valid SpatiallyBinnedSpectra object!')
        if binned_spectra.hdu is None:
            raise ValueError('Provided SpatiallyBinnedSpectra object is undefined!')
        self.binned_spectra = binned_spectra

        # Check stellar-continuum model object, if provided
        self.stellar_continuum = None
        if stellar_continuum is not None:
            if not isinstance(stellar_continuum, StellarContinuumModel):
                raise TypeError('Provided stellar continuum must have StellarContinuumModel type!')
            if stellar_continuum.hdu is None:
                raise ValueError('Provided StellarContinuumModel is undefined!')
            self.stellar_continuum = stellar_continuum

        # Check emission-line model object, if provided
        self.emission_line_model = None
        if emission_line_model is not None:
#            if not isinstance(emission_line_model, EmissionLineModel):
#                raise TypeError('Provided emission line models must be of type EmissionLineModel.')
            if emission_line_model.hdu is None:
                raise ValueError('Provided EmissionLineModel is undefined!')
            self.emission_line_model = emission_line_model

        # What stellar continuum is available?
        #  - Assume the stellar continuum can always be extracted if a
        #    StellarContinuumModel object is provided
        #  - Check if the EmissionLineModel fitter has the appropriate
        #    function; True for Sasuke, not currently true for Elric
        eml_stellar_continuum_available = self.emission_line_model is not None \
                and callable(self.emission_line_model.method['fitclass'].construct_continuum_models)

        # What spectra to use?
        #  - Assume StellarContinuumModel always fits the binned spectra
        #  - The EmissionLineModel fits the binned spectra or unbinned
        #    spaxels as specified by its deconstruct_bins flag
        measure_on_unbinned_spaxels = self.emission_line_model is not None \
                and self.emission_line_model.method['fitpar']['deconstruct_bins'] != 'ignore'

        self.spatial_shape =self.binned_spectra.spatial_shape
        self.nspec = self.binned_spectra.cube.nspec if measure_on_unbinned_spaxels \
                            else self.binned_spectra.nbins
        self.spatial_index = self.binned_spectra.spatial_index.copy()
        
        #---------------------------------------------------------------
        # Get the good spectra
        good_snr = self._flag_good_spectra(measure_on_unbinned_spaxels)

        # Set the number of bins measured and missing bins
        self.nbins = numpy.sum(good_snr)
        self.missing_bins = [] if measure_on_unbinned_spaxels else self._get_missing_bins()

        # Get the redshifts to apply
        self._assign_redshifts(redshift, measure_on_unbinned_spaxels, good_snr)

        # Report
        if not self.quiet:
            log_output(self.loggers, 1, logging.INFO, '-'*50)
            log_output(loggers, 1, logging.INFO, f'{"EMISSION-LINE MOMENTS":^50}')
            log_output(self.loggers, 1, logging.INFO, '-'*50)
            log_output(self.loggers, 1, logging.INFO, '-'*50)
            log_output(self.loggers, 1, logging.INFO, 'Measurements for {0}'.format(
                        'unbinned spaxels' if measure_on_unbinned_spaxels else 'binned spectra'))
            if self.stellar_continuum is None and self.emission_line_model is None:
                log_output(self.loggers, 1, logging.INFO, 'No stellar continuum available.')
            elif self.emission_line_model is not None and eml_stellar_continuum_available:
                log_output(self.loggers, 1, logging.INFO,
                           'Using stellar continuum from emission-line model fit.')
            else:
                log_output(self.loggers, 1, logging.INFO,
                           'Using stellar continuum from stellar-continuum model fit.')
            log_output(self.loggers, 1, logging.INFO, f'Number of spectra: {self.nspec}')
            if not measure_on_unbinned_spaxels and len(self.binned_spectra.missing_bins) > 0:
                log_output(self.loggers, 1, logging.INFO,
                           f'Missing bins: {len(self.binned_spectra.missing_bins)}')
            log_output(self.loggers, 1, logging.INFO,
                       f'With good S/N and to measure: {numpy.sum(good_snr)}')

        # Make sure there are good spectra
        if numpy.sum(good_snr) == 0:
            raise ValueError('No good spectra for measurements!')

        #---------------------------------------------------------------
        # (Re)Set the output paths
        self.directory_path, self.output_file = \
                EmissionLineMoments.default_paths(self.binned_spectra.cube, self.database['key'],
                        self.binned_spectra.rdxqa.method['key'], self.binned_spectra.method['key'],
                        stelcont_method=None if self.stellar_continuum is None 
                                        else self.stellar_continuum.method['key'],
                        elmodel_method=None if self.emission_line_model is None 
                                       else self.emission_line_model.method['key'],
                        output_path=output_path, output_file=output_file)
#        self._set_paths(directory_path, dapver, analysis_path, output_file)

        _overwrite = self.database['overwrite'] if overwrite is None else overwrite

        #---------------------------------------------------------------
        # Check that the file path is defined
        ofile = self.file_path()
        if ofile is None:
            raise ValueError('File path for output file is undefined!')

        # Report
        if not self.quiet:
            log_output(self.loggers,1,logging.INFO, f'Output path: {self.directory_path}')
            log_output(self.loggers,1,logging.INFO, f'Output file: {self.output_file}')
        
        #---------------------------------------------------------------
        # If the file already exists, and not overwriting, just read the
        # file
        if ofile.exists() and not _overwrite:
            self.hardcopy = True
            if not self.quiet:
                log_output(self.loggers, 1, logging.INFO, 'Using existing file')
            self.read(checksum=self.checksum)
            if not self.quiet:
                log_output(self.loggers, 1, logging.INFO, '-'*50)
            return

        #---------------------------------------------------------------
        # Get the spectra to use for the measurements
        wave, flux, ivar, sres \
                = EmissionLineFit.get_spectra_to_fit(self.binned_spectra,
                                                     pixelmask=self.database['pixelmask'],
                                                     select=good_snr,
                                                     original_spaxels=measure_on_unbinned_spaxels)

        # Get the continuum if there is any
        if eml_stellar_continuum_available:
            if measure_on_unbinned_spaxels:
                binid = numpy.full(self.binned_spectra.spatial_shape, -1, dtype=int)
                binid.ravel()[good_snr] = numpy.arange(self.nbins)
                continuum = self.emission_line_model.fill_continuum_to_match(binid)
            else:
                continuum = self.emission_line_model.fill_continuum_to_match(
                                                self.binned_spectra['BINID'].data,
                                                missing=self.binned_spectra.missing_bins)
        elif self.stellar_continuum is not None:
            continuum = self.stellar_continuum.fill_to_match(self.binned_spectra['BINID'].data,
                                                        missing=self.binned_spectra.missing_bins)

        #---------------------------------------------------------------
        # Perform the moment measurements
        _redshift = self.redshift[good_snr]
        measurements = self.measure_moments(self.database.momdb, wave, flux, ivar=ivar, sres=sres,
                                            continuum=continuum, redshift=_redshift,
                                            bitmask=self.bitmask)

        #---------------------------------------------------------------
        # Perform the equivalent width measurements
        include_band = numpy.array([numpy.invert(self.database.momdb.dummy)]*self.nbins) \
                        & numpy.invert(self.bitmask.flagged(measurements['MASK'],
                                                            flag=['BLUE_EMPTY', 'RED_EMPTY']))

        # Set the line center at the center of the primary passband
        line_center = (1.0+_redshift)[:,None]*self.database.momdb['restwave'][None,:]
        measurements['BMED'], measurements['RMED'], pos, measurements['EWCONT'], \
                measurements['EW'], measurements['EWERR'] \
                        = emission_line_equivalent_width(wave, flux,
                                                         self.database.momdb['blueside'],
                                                         self.database.momdb['redside'],
                                                         line_center, measurements['FLUX'],
                                                         redshift=_redshift,
                                                         line_flux_err=measurements['FLUXERR'],
                                                         include_band=include_band)

        # Flag non-positive measurements
        measurements['MASK'][include_band & numpy.invert(pos)] \
                    = self.bitmask.turn_on(measurements['MASK'][include_band & numpy.invert(pos)],
                                           'NON_POSITIVE_CONTINUUM')

        #---------------------------------------------------------------
        # Set the number of bins measured, missing bins, and bin IDs
        measurements['BINID'] = numpy.arange(self.nbins) if measure_on_unbinned_spaxels \
                                    else self.binned_spectra['BINS'].data['BINID'][good_snr]
        measurements['BINID_INDEX'] = numpy.arange(self.nbins) if measure_on_unbinned_spaxels \
                                        else numpy.arange(self.binned_spectra.nbins)[good_snr]
        measurements_binid = None
        if measure_on_unbinned_spaxels:
            measurements_binid = numpy.full(self.binned_spectra.spatial_shape, -1, dtype=int)
            measurements_binid.ravel()[good_snr] = numpy.arange(self.nbins)

        #---------------------------------------------------------------
        # Initialize the header
        self.hardcopy = hardcopy
        pri_hdr = self._initialize_primary_header(measurements_binid=measurements_binid)
        map_hdr = DAPFitsUtil.build_map_header(self.binned_spectra.cube.fluxhdr,
                                               'K Westfall <westfall@ucolick.org>')

        # Get the spatial map mask
        map_mask = numpy.zeros(self.spatial_shape, dtype=self.bitmask.minimum_dtype())

        # Account for measurements on individual spaxels
        if measurements_binid is None:
            # Add any spaxel not used because it was flagged by the
            # binning step
            indx = self.binned_spectra['MAPMASK'].data > 0
            map_mask[indx] = self.bitmask.turn_on(map_mask[indx], 'DIDNOTUSE')
            # Isolate any spaxels with foreground stars
            indx = self.binned_spectra.bitmask.flagged(self.binned_spectra['MAPMASK'].data,
                                                       'FORESTAR')
            map_mask[indx] = self.bitmask.turn_on(map_mask[indx], 'FORESTAR')
            # Get the bins that were below the S/N limit
            indx = numpy.invert(DAPFitsUtil.reconstruct_map(self.spatial_shape,
                                                        self.binned_spectra['BINID'].data.ravel(),
                                                        good_snr, dtype='bool')) & (map_mask == 0)
            map_mask[indx] = self.bitmask.turn_on(map_mask[indx], 'LOW_SNR')

            # Get the bin ids with measured indices
            bin_indx = DAPFitsUtil.downselect_bins(self.binned_spectra['BINID'].data.ravel(),
                                                measurements['BINID']).reshape(self.spatial_shape)
        else:
            # Assume any model with a binid less than zero is from a
            # spaxel that was not used
            indx = measurements_binid < 0
            map_mask[indx] = self.bitmask.turn_on(map_mask[indx], 'DIDNOTUSE')

            # The number of valid bins MUST match the number of
            # measurements
            nvalid = numpy.sum(numpy.invert(indx))
            if nvalid != len(measurements):
                raise ValueError('Provided id does not match the number of measurements.')

            # Get the bin ids with fitted models
            bin_indx = measurements_binid

        # Save the data to the hdu attribute
        self.hdu = fits.HDUList([fits.PrimaryHDU(header=pri_hdr),
                                 fits.ImageHDU(data=bin_indx, header=map_hdr, name='BINID'),
                                 fits.ImageHDU(data=map_mask, header=map_hdr, name='MAPMASK'),
                                 self.database.momdb.to_datatable().to_hdu(name='ELMBAND'),
                                 measurements.to_hdu(name='ELMMNTS')])

        #---------------------------------------------------------------
        # Write the data, if requested
        if self.hardcopy:
            if not self.directory_path.exists():
                self.diretory_path.mkdir(parents=True)
            self.write(overwrite=_overwrite)
        if not self.quiet:
            log_output(self.loggers, 1, logging.INFO, '-'*50)

    def write(self, overwrite=False):
        """
        Write the hdu object to the file.
        """
        DAPFitsUtil.write(self.hdu, str(self.file_path()), overwrite=overwrite, checksum=True,
                          loggers=self.loggers, quiet=self.quiet)

    def read(self, ifile=None, strict=True, checksum=False):
        """

        Read an existing file with a previously binned set of spectra.

        """
        _ifile = self.file_path() if ifile is None else Path(ifile).resolve()
        if not _ifile.exists():
            raise FileNotFoundError('File does not exist!: {0}'.format(ifile))

        if self.hdu is not None:
            self.hdu.close()

        self.hdu = DAPFitsUtil.read(str(_ifile), checksum=checksum)

        # Confirm that the internal method is the same as the method
        # that was used in writing the file
        if self.hdu['PRIMARY'].header['ELMKEY'] != self.database['key']:
            if strict:
                raise ValueError('Keywords in header do not match specified method keywords!')
            else:
                warnings.warn('Keywords in header do not match specified method keywords!')
        # TODO: "strict" should also check other aspects of the file to
        # make sure that the details of the method are also the same,
        # not just the keyword

        self.nbins = self.hdu['PRIMARY'].header['NBINS']
        unique_bins = numpy.unique(self.hdu['BINID'].data.ravel()) \
                            if self.hdu['PRIMARY'].header['ELMREBIN'] else None
        self.missing_bins = self._get_missing_bins(unique_bins=unique_bins)

    # TODO: Use the EmissionMomentsDB.channel_names function!
    def channel_names(self):
        return numpy.array([ '{0}-{1}'.format(n,int(w)) \
                        for n,w in zip(self['ELMBAND'].data['NAME'],
                                       self['ELMBAND'].data['RESTWAVE']) ])




# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
r"""
Classes to handle MaNGA DAP analysis plans.

----

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""

from pathlib import Path
from copy import deepcopy

from IPython import embed

import tomli

from ..par.util import recursive_dict_str_to_None
from . import defaults

from mangadap.proc.reductionassessments import ReductionAssessmentDef
from mangadap.proc.spatiallybinnedspectra import SpatiallyBinnedSpectraDef
from mangadap.proc.stellarcontinuummodel import StellarContinuumModelDef
from mangadap.proc.emissionlinemoments import EmissionLineMomentsDef
from mangadap.proc.emissionlinemodel import EmissionLineModelDef
from mangadap.proc.spectralindices import SpectralIndicesDef


class AnalysisPlan:
    """
    Container class for a set of analysis plans.
    """
    def __init__(self, plan, cube=None, analysis_path=None):
        if not isinstance(plan, dict):
            raise TypeError('Plan must be provided as a dictionary.')
        # Copy the provided plan dictionary so that it can be changed/filled
        # throughout the execution of the code.
        self.plan = deepcopy(plan)
        self.plan_keys = list(self.plan.keys())
        self.nplans = len(self.plan_keys)
        # Make sure that the plans have a key name.  If not, just use the
        # dictionary key of the dictionary for each plan.
        for key in self.plan_keys:
            if 'key' not in self.plan[key]:
                self.plan['key'] = key
        self.analysis_path = Path('.' if analysis_path is None else analysis_path).resolve()
        self.cube = cube
        self._validate()
        self.parse()

    def _validate(self):
        """
        Validate the provided plans.
        """
        # For now, just make sure that all the plans have the base top-level keys
        required_keys = ['rdxqa', 'binning', 'continuum', 'eline_moments', 'eline_fits', 'indices']
        for key in self.plan_keys:
            for method_key in required_keys:
                if method_key not in self[key].keys():
                    self.plan[key][method_key] = {}

        # Recursively convert None strings into None types
        self.plan = recursive_dict_str_to_None(self.plan)

        # TODO: Cross validate parameter sets between plans!!

    def __getitem__(self, key):
        """
        Return the value of the designated key.
        """
        return self.plan[key]

    def keys(self):
        return self.plan.keys()

    @classmethod
    def from_toml(cls, ifile, **kwargs):
        """
        Instantiate the plan from a TOML file.
        """
        _ifile = Path(ifile).resolve()
        if not _ifile.exists():
            raise FileNotFoundError(f'{_ifile} does not exist!')

        with open(_ifile, 'rb') as f:
            plan = tomli.load(f)

        return cls(plan, **kwargs)

    @classmethod
    def default(cls, **kwargs):
        """
        Return the default analysis plan set.
        """
        f = defaults.dap_config_root() / 'default_plan.toml'
        return cls.from_toml(f, **kwargs)

    def parse(self):
        self.rdxqa = {key: None if self[key]['rdxqa'] is None else
                        ReductionAssessmentDef.from_dict(self[key]['rdxqa'])
                        for key in self.plan.keys()}
        self.binning = {key: None if self[key]['binning'] is None else
                            SpatiallyBinnedSpectraDef.from_dict(self[key]['binning'])
                            for key in self.plan.keys()}
        self.continuum = {key: None if self[key]['continuum'] is None else
                            StellarContinuumModelDef.from_dict(self[key]['continuum'])
                            for key in self.plan.keys()}
        self.elmom = {key: None if self[key]['eline_moments'] is None else
                        EmissionLineMomentsDef.from_dict(self[key]['eline_moments']) 
                        for key in self.plan.keys()}
        self.elfit = {key: None if self[key]['eline_fits'] is None else
                        EmissionLineModelDef.from_dict(self[key]['eline_fits'])
                        for key in self.plan.keys()}
        self.sindx = {key: None if self[key]['indices'] is None else
                        SpectralIndicesDef.from_dict(self[key]['indices'])
                        for key in self.plan.keys()}

    def common_path(self):
        """
        Return the path for data common to all plans.

        Args:
            cube (:class:`~mangadap.datacube.datacube.DataCube`, optional):
                Cube being analyzed.  Passed for cube-specific path
                specification.  Not used by this base class.

        Returns:
            `Path`_: Path object for the "common" output
        """
        return self.analysis_path / 'common'

    def method_path(self, plan_index=0, qa=False, ref=False):
        """
        Return the path for method-specific output.

        Args:
            cube (:class:`~mangadap.datacube.datacube.DataCube`, optional):
                Cube being analyzed.  Passed for cube-specific path
                specification.  Not used by this base class.
            plan_index (:obj:`int`, optional):
                The index of the plan.  This is used to select the 'key' of the
                analysis plan being used, which is used as the subdirectory for
                the output.
            qa (:obj:`bool`, optional):
                Flag that the output is specifically quality assessment plots.
            ref (:obj:`bool`, optional):
                Flag that the output is specifically reference data.

        Returns:
            `Path`_: Path object for the method-specific output

        Raises:
            ValueError:
                Raised if the plan index is invalid or if both qa and ref are true.

        """
        if plan_index < 0 or plan_index >= self.nplans:
            raise ValueError(f'Invalid index ({plan_index}); 0 <= index < {self.nplans}.')
        if qa and ref:
            raise ValueError('Cannot provide path for both qa and ref directory.  Pick one.')
        root = self.analysis_path / self.plan[self.plan_keys[plan_index]]['key']
        if not qa and not ref:
            return root
        if qa:
            return root / 'qa'
        return root / 'ref'

    # TODO: Restrict mode?
    def dap_file_root(self, cube, mode=None, plan_index=None):
        """
        Return the general root name for an output file.

        The default returned by this base class is::

        {cube.output_root}-{mode}-{self['key'][plan_index]}
        
        where ``mode`` and ``self['key'][plan_index]`` are only included if
        ``mode`` or ``plan_index`` are provided.

        Args:
            cube (:class:`~mangadap.datacube.datacube.DataCube`):
                Cube being analyzed.  This is used to provide the primary root
                of the output name.
            mode (:obj:`str`, optional):
                An optional "mode" included in the file name.  If None, not
                include in the output string.
            plan_index (:obj:`int`, optional):
                The index of the plan.  This is used to select the 'key' of the
                analysis plan being used, which is included in the file name.
                If None, this is not included in the file name.
    
        Returns:
            :obj:`str`: General root for output DAP file names.
        """
        root = cube.output_root
        if mode is not None:
            root = f'{root}-{mode}'
        if plan_index is not None:
            if plan_index < 0 or plan_index >= self.nplans:
                raise ValueError(f'Invalid index ({plan_index}): 0 <= index < {self.nplans}.')
            plan_name = list(self.keys())[plan_index]
            root = f'{root}-{self[plan_name]["key"]}'
        return root







import os
import warnings

import pytest

from mangadap.config import defaults

def data_test_file(filename=None):
    root = os.path.join(defaults.dap_data_root(), 'tests')
    return root if filename is None else os.path.join(root, filename)

def remote_data_file(filename=None):
    root = os.path.join(defaults.dap_data_root(), 'remote')
    return root if filename is None else os.path.join(root, filename)

def remote_data_files():
    return ['manga-7815-3702-LINCUBE.fits.gz', 'manga-7815-3702-LINRSS.fits.gz',
            'manga-7815-3702-LOGCUBE.fits.gz', 'manga-7815-3702-LOGRSS.fits.gz']
#    return ['manga-7815-6101-LINCUBE.fits.gz', 'manga-7815-6101-LINRSS.fits.gz',
#            'manga-7815-6101-LOGCUBE.fits.gz', 'manga-7815-6101-LOGRSS.fits.gz']

drp_test_version = 'v3_1_1'
dap_test_version = '3.1.0'
remote_available = all([os.path.isfile(remote_data_file(f)) for f in remote_data_files()])
drpcomplete_available = os.path.isfile(remote_data_file('drpcomplete_{0}.fits'.format(
                                       drp_test_version)))
drpall_available = os.path.isfile(remote_data_file('drpall-{0}.fits'.format(drp_test_version)))

requires_remote = pytest.mark.skipif(not remote_available, reason='Remote data files are missing.')
requires_drpcomplete = pytest.mark.skipif(not drpcomplete_available,
                                          reason='Remote DRPComplete is missing.')
requires_drpall = pytest.mark.skipif(not drpall_available, reason='Remote DRPall is missing.')

if not remote_available:
    warnings.warn('Remote data not available.  Some tests will be skipped.')
if not drpcomplete_available:
    warnings.warn('Remote DRPComplete not available.  Some tests will be skipped.')
if not drpall_available:
    warnings.warn('Remote DRPall not available.  Some tests will be skipped.')




import os
from ..version import version

def get_environ():

    # Default versions
    _MANGADRP_VER = 'v3_1_1'
    _MANGACORE_VER = 'v1_9_1'

    # First set the default environment
    environ = {'MANGADRP_VER': _MANGADRP_VER,
               'MANGA_SPECTRO_REDUX': os.path.join(os.getcwd(), 'MaNGA', 'redux'),
               'MANGADAP_VER': version,
               'MANGA_SPECTRO_ANALYSIS': os.path.join(os.getcwd(), 'MaNGA', 'analysis'),
               'MANGACORE_VER': _MANGACORE_VER,
               'MANGACORE_DIR': os.path.join(os.getcwd(), 'MaNGA', 'core', _MANGACORE_VER)}

    # Replace with variables from working environment if they exist
    for key in environ.keys():
        if key in os.environ:
            environ[key] = os.environ[key]

    return environ

manga_environ = get_environ()

#from .manga import MaNGAAnalysisPlan




# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
Defines a class used to produce and provide an interface for data
required to create :ref:`execution-config`.

The drpcomplete file/data is primarily required for the survey-level
:ref:`execution` of the DAP. However, it does provide useful information
regarding the completed DRP files.

----

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""

import os
import time
import warnings
import glob


from IPython import embed

import numpy

from astropy.io import fits
import astropy.constants

from pydl.pydlutils.yanny import yanny

from ..datacube import MaNGADataCube
from ..spectra import MaNGARSS
#from ..config import defaults
from ..config import manga
from ..util.parser import arginp_to_list, list_to_csl_string, parse_drp_file_name

class DRPComplete:
    r"""
    Database with information needed by the DAP to analyze the completed
    DRP files.

    This class searches the defined paths for files resulting from the
    3D phase of the MaNGA DRP, and then collates the information
    necessary to run those files through the MaNGA DAP.  The necessary
    parameters are pulled from the provided platetargets files; see
    :func:`update`.

    Args:

        platelist (:obj:`str`, :obj:`list`, optional):
            List of plates to search for.  Default is to search the full
            DRP path.
        ifudesignlist (:obj:`str`, :obj:`list`, optional):
            List of ifudesign to search for.  Default is to search the
            full DRP path.
        drpall (:obj:`str`, optional):
            The full path to the DRPall fits file.  Default is set by
            :func:`mangadap.config.defaults.drpall_file`.
        platetargets (:obj:`str`, :obj:`list`, optional):
            List of platetargets files to search through to find any
            given plate ifudesign combination.  Default is returned as
            the first element in
            :func:`mangadap.config.defaults.plate_target_files`.
        catid (:obj:`str`, :obj:`list`, optional):
            List of target catalog ID numbers.  Default is returned as
            the second element in
            :func:`mangadap.config.defaults.plate_target_files`.
        drpver (:obj:`str`, optional):
            DRP version, which is:

                - used to define the default DRP redux path
                - used when declaring a drpfits instance
                - used in the name of the drpcomplete fits file
                - included as a header keyword in the output file

            Default is defined by
            :func:`mangadap.config.defaults.drp_version`.
        redux_path (:obj:`str`, optional): 
            The path to the top level directory containing the DRP
            output files; this is the same as the ``redux_path`` in
            the :class:`mangadap.util.drpfits.DRPFits` class. Default
            is defined by
            :func:`mangadap.config.defaults.drp_redux_path`.
        dapver (:obj:`str`, optional):
            DAP version, which is:

                - used to define the default DAP analysis path
                - included as a header keyword in the output drpcomplete
                  fits file

            Default is defined by
            :func:`mangadap.config.defaults.dap_version`.
        analysis_path (:obj:`str`, optional):
            The path to the top level directory for the DAP output
            files; this is **different** from the directory_path in the
            :class:`mangadap.dapfile` class.  Default is defined by
            :func:`mangadap.config.defaults.dap_analysis_path`
        directory_path (:obj:`str`, optional):
            Direct path to the output file produced using
            :func:`mangadap.config.defaults.dap_common_path`
        readonly (:obj:`bool`, optional):
            Flag that the drpcomplete fits file is only opened for
            reading, not for updating.  If True, many of the attributes
            will be set to None.

    Attributes:
        platelist (:obj:`list`):
            List of plates to search for, see above.
        ifudesignlist (:obj:`list`):
            List of IFU designs to search for, see above.
        drpall (:obj:`str`):
            The DRPall file, see above.
        platetargets (`numpy.ndarray`):
            List of platetargets files to search through, see above.
        catid (`numpy.ndarray`):
            List of target catalog IDs, see above
        drpver (:obj:`str`):
            DRP version, see above.
        redux_path (:obj:`str`):
            Path to the top level of directory for the DRP output, see
            above.
        dapver (:obj:`str`):
            DAP version, see above.
        analysis_path (:obj:`str`):
            Path to the top level directory for the DAP output files,
            see above.
        directory_path (:obj:`str`):
            Direct path to the output file produced using
            :func:`mangadap.config.defaults.dap_common_path`
        readonly (:obj:`bool`):
            Flag that the drpcomplete fits file is only opened for
            reading, not for updating.
        hdu (`astropy.io.fits.HDUList`_):
            Fits data with binary table data.
        nobs (:obj:`int`):
            Number of observations in the file

    """
    def __init__(self, platelist=None, ifudesignlist=None, drpall=None, platetargets=None,
                 catid=None, drpver=None, redux_path=None, dapver=None, analysis_path=None,
                 directory_path=None, readonly=False):

        # Input properties
        self.drpver = manga.drp_version() if drpver is None else str(drpver)
        self.redux_path = manga.drp_redux_path(self.drpver) \
                                if redux_path is None else str(redux_path)

        self.dapver = manga.dap_version() if dapver is None else str(dapver)
        self.analysis_path = manga.dap_analysis_path(self.drpver, self.dapver) \
                             if analysis_path is None else str(analysis_path)
        self.directory_path = self.analysis_path / 'common' if directory_path is None \
                                    else str(directory_path)

        self.hdu = None
        self.nobs = None
        if os.path.exists(self.file_path()):
            print('READING')
            self._read()

        if readonly:
            self.readonly=True
            self.platelist = None
            self.ifudesignlist = None
            self.platetargets = None
            self.catid = None
            return

        self.readonly = False

        self.platelist = arginp_to_list(platelist, evaluate=True)
        self.ifudesignlist = arginp_to_list(ifudesignlist, evaluate=True)

        self.drpall = manga.drpall_file(drpver=self.drpver,redux_path=self.redux_path) \
                            if drpall is None else drpall
        
        if (platetargets is not None and catid is None) or \
           (platetargets is None and catid is not None):
            raise ValueError('To use user-provided platetargets files, must provide both '
                             'platetargets and catid.')

        self.platetargets = None
        if platetargets is not None:
            self.platetargets = numpy.array( arginp_to_list(platetargets) )
            self.catid = numpy.array( arginp_to_list(catid) ).astype(int)
        else:
            try:
                self.platetargets, self.catid = manga.plate_target_files()
            except:
                warnings.warn('Could not define platetargets files.  '
                              'Updates must use DRPall file.')

    def __getitem__(self, key):
        return self.hdu['DRPC'].data[key]

    def __len__(self):
        return self.nobs

    # ******************************************************************
    #  Utility functions
    # ******************************************************************

    def _read_platetargets(self):
        """
        Read all the platetargets files using `pydl.pydlutils.yanny`_
        and return a list of yanny structures.

        Returns:
            list: A list of yanny structures, one per platetargets file.

        Raises:
            FileNotFoundError: Raised if cannot open one or more
                platetargets files.

        .. todo::
            This should be made more efficient by collating the required
            plateTargets data into a single record array!
        """
        plttrg_data = []

        root = '/'+os.path.join(*(self.platetargets[0].split('/')[:-1]))
        print('plateTargets root directory: {0}'.format(root))

        for p in self.platetargets:
            pfile = p.split('/')[-1]
            print('Reading plateTargets file: {0}'.format(pfile), end='\r')
            # Check that the file exists
            if not os.path.exists(p):
                raise FileNotFoundError('Cannot open {0}!'.format(p))
            # Read and append the data
            plttrg_data.append( yanny(filename=p) )
        print('\nReading plateTargets file: DONE')

        return plttrg_data

    def _read_fix_data(self, fix_file):
        """
        Read fix data from the provided file.

        Args:
            fix_file (:obj:`str`):
                SDSS parameter file with the fix data.

        Returns:
            yanny: A yanny structure with the data fixes.  Returns
            None and raises a warning if the file does not exist.
        """
        if not os.path.isfile(fix_file):
            warnings.warn('No redshift fix file available.')
            return None
        return yanny(filename=fix_file)

    def _read_redshift_fix(self):
        """
        Wrapper for :func:`_read_fix_data` and
        :func:`~mangadap.config.defaults.redshift_fix_file` that
        returns the redshift fix data.
        """
        return self._read_fix_data(manga.redshift_fix_file())

    def _read_photometry_fix(self):
        """
        Wrapper for :func:`_read_fix_data` and
        :func:`~mangadap.config.defaults.photometry_fix_file` that
        returns the photometry fix data.
        """
        return self._read_fix_data(manga.photometry_fix_file())

    def _match_platetargets(self, quiet=True):
        """
        Read the platetargets files and match the data therein to the
        completed DRP files based on the plate and ifudesign.

        If a plate-ifudesign combination is not found in the
        plateTargets file, the other parameter values are set to -9999
        and the MaNGA ID is set to NULL.

        If the plate-ifudesign combination is found, the columns in the
        plateTargets files that the code expects to find are 'plateid',
        'ifudesign', 'mangaid', 'object_ra', and 'object_dec'; the
        values that will be replaced with 'NULL' (str) or -9999 (int or
        float) if they do not exist are 'nsa_version', 'nsa_nsaid',
        'manga_target1', 'manga_target3', 'z' or 'nsa_z',
        'nsa_elpetro_ba', 'nsa_elpetro_phi', 'nsa_elpetro_th50_R',
        'nsa_vdisp'.
        
        .. todo::

            - Instead of searching through all files for the correct
              plate-ifudesign, use the MaNGA ID to get the correct
              catalog?  Requires getting the MaNGA ID from somewhere,
              but it takes a long time to read from the fits files...
            - Is there some more concise method of performing the same
              thing as what is done below with many try/except blocks?

        Args:
            quiet (:obj:`bool`, optional):
                Suppress terminal output

        Returns:
            numpy.array: 14 arrays with: MaNGA ID, object right
            ascension, object declination, catalog ID, index of the
            entry in the catalog, catalog version, ID of object in the
            catalog, main MaNGA survey target bitmask, ancillary MaNGA
            survey target bitmask, velocity, velocity dispersion,
            ellipticity, position angle, and effective radius.
        """
        # Read the platetargets file
        plttrg_data = self._read_platetargets()
        ntrg = len(plttrg_data)
        print('Read {0} plateTargets file(s)'.format(ntrg))

        # Initialize the output arrays (needed in case some DRP targets not found)
        n_drp = len(self.platelist)
        mangaid = []
        # TODO: try mangaid = numpy.empty(n_drp, dtype=object) ?
        objra = numpy.full(n_drp, -9999.0, dtype=numpy.float64)
        objdec = numpy.full(n_drp, -9999.0, dtype=numpy.float64)
        catid = numpy.full(n_drp, -9999, dtype=numpy.int32)
        catindx = numpy.full(n_drp, -9999, dtype=numpy.int32)
        trg_version = []
        # TODO: try trg_version = numpy.empty(n_drp, dtype=object) ?
        trg_id = numpy.full(n_drp, -9999, dtype=numpy.int32)
        manga_trg1 = numpy.full(n_drp, -9999, dtype=numpy.int64)
        manga_trg3 = numpy.full(n_drp, -9999, dtype=numpy.int64)
        vel = numpy.full(n_drp, -9999.0, dtype=numpy.float64)
        veldisp = numpy.full(n_drp, -9999.0, dtype=numpy.float64)
        ell = numpy.full(n_drp, -9999.0, dtype=numpy.float64)
        pa = numpy.full(n_drp, -9999.0, dtype=numpy.float64)
        Reff = numpy.full(n_drp, -9999.0, dtype=numpy.float64)

        print('Searching platetargets file for observed galaxies...', end='\r')
        for i in range(n_drp):
            plttrg_j = 0
            mangaid_i = 'NULL'
            for j in range(ntrg):
                indx = numpy.where((plttrg_data[j]['PLTTRGT']['plateid'] == self.platelist[i]) &
                                (plttrg_data[j]['PLTTRGT']['ifudesign'] == self.ifudesignlist[i]))
                if len(indx[0]) > 1:
                    raise ValueError('Multiple instances of {0}-{1} (MaNGA ID={2}) in {3}'.format(
                                     self.platelist[i], self.ifudesignlist[i], mangaid_i,
                                     self.platetargets[j]))
                if len(indx[0]) == 1:
                    plttrg_j = j
                    mangaid_i = plttrg_data[j]['PLTTRGT']['mangaid'][indx][0].decode("ascii")
                    if not quiet:
                        print('Found {0}-{1} (MaNGA ID={2}) in {3} (CAT={4})'.format(
                                        self.platelist[i], self.ifudesignlist[i], mangaid_i,
                                        self.platetargets[j], self.catid[j]))
                    break

            if len(indx[0]) == 0:
                warnings.warn('Could not find {0}-{1} in any plateTargets file!'.format(
                                                        self.platelist[i], self.ifudesignlist[i]))
                mangaid = mangaid + ['NULL']
                continue
            mangaid.append(plttrg_data[plttrg_j]['PLTTRGT']['mangaid'][indx][0].decode("ascii"))
            objra[i] = plttrg_data[plttrg_j]['PLTTRGT']['object_ra'][indx][0]
            objdec[i] = plttrg_data[plttrg_j]['PLTTRGT']['object_dec'][indx][0]

            # From David Wake:
            #
            # "MANGAID consists of CATID-CATIND, where CATID identifies
            # a parent catalog, in the case of the main manga samples
            # CATID = 1 which refers to nsa_v1_0_0.fits, and CATIND is
            # the position within that catalog (zero indexed). So if you
            # strip out CATIND from MANGAID you have the position within
            # nsa_v1_0_0.fits without any matching required."
            catid[i], catindx[i] = map(lambda x: int(x), mangaid[i].split('-'))
            if catid[i] != self.catid[plttrg_j]:
                warnings.warn('{0}-{1} (MaNGA ID={2}) found in {3} (CAT={4})!'.format(
                                self.platelist[i], self.ifudesignlist[i], mangaid_i,
                                self.platetargets[j], self.catid[j]))

            try:
                trg_version.append(plttrg_data[plttrg_j]['PLTTRGT']['nsa_version'][indx][0].decode(
                                                                                        'ascii'))
            except:
                trg_version.append('NULL')

            try:
                trg_id[i] = plttrg_data[plttrg_j]['PLTTRGT']['nsa_nsaid'][indx][0]
            except:
                trg_id[i] = -9999
                
            try:
                manga_trg1[i] = plttrg_data[plttrg_j]['PLTTRGT']['manga_target1'][indx][0]
            except:
                manga_trg1[i] = -9999

            try:
                manga_trg3[i] = plttrg_data[plttrg_j]['PLTTRGT']['manga_target3'][indx][0]
            except:
                manga_trg3[i] = -9999

            # As of MPL-6, platetargets files now include the redshift
            # from the targeting catalog which is the combination of the
            # NSA data and the ancillary targets; the NSA only redshift
            # column is 'nsa_z'.  To be compatible with previous
            # versions, first try to grab the redshift using the keyword
            # 'z', then try with 'nsa_z', then just set it to -9999.
            try:
                vel[i] = plttrg_data[plttrg_j]['PLTTRGT']['z'][indx][0] \
                            * astropy.constants.c.to('km/s').value
            except:
                try:
                    vel[i] = plttrg_data[plttrg_j]['PLTTRGT']['nsa_z'][indx][0] \
                                * astropy.constants.c.to('km/s').value
                except:
                    vel[i] = -9999.0

            try:
                if plttrg_data[plttrg_j]['PLTTRGT']['nsa_elpetro_ba'][indx][0] < 0:
                    raise
                ell[i] = 1.0-plttrg_data[plttrg_j]['PLTTRGT']['nsa_elpetro_ba'][indx][0]
            except:
                ell[i] = -9999.0
                warnings.warn('Ellipticity for {0}-{1} is bogus!'.format(self.platelist[i],
                                                                         self.ifudesignlist[i]))

            try:
                # THERE'S A BUG IN THE MANGACORE/V1_2_0 PLATETARGETS FILES
                pa[i] = plttrg_data[plttrg_j]['PLTTRGT']['nsa_elpetro_phi'][indx][0]
                if pa[i] < 0:
                    raise
            except:
                pa[i] = -9999.0
                warnings.warn('PA for {0}-{1} is bogus!'.format(self.platelist[i],
                                                                self.ifudesignlist[i]))

            try:
                Reff[i] = plttrg_data[plttrg_j]['PLTTRGT']['nsa_elpetro_th50_r'][indx][0]
            except:
                Reff[i] = -9999.0

            try:
                veldisp[i] = plttrg_data[plttrg_j]['PLTTRGT']['nsa_vdisp'][indx][0]
            except:
                veldisp[i] = -9999.0

            # Correct for known nan issue
            if numpy.isnan(ell[i]):
                raise ValueError('nan encountered!')
            if numpy.isnan(pa[i]):
                raise ValueError('nan encountered!')

        print('Searching platetargets file for observed galaxies...DONE')

        return numpy.array(mangaid), objra, objdec, catid, catindx, numpy.array(trg_version), \
               trg_id, manga_trg1, manga_trg3, vel, veldisp, ell, pa, Reff

    def _match_drpall(self):
        """
        Find the data needed for the DRP complete databas in the DRPall
        file.

        If a plate-ifudesign combination is not found in the DRPall
        file, the other parameter values are set to -9999 and the MaNGA
        ID is set to NULL.

        If the plate-ifudesign combination is found, the columns in the
        plateTargets files that the code expects to find are 'plate',
        'ifudsgn', 'mangaid', 'objra', 'objdec'; the values that will be
        replaced with 'NULL' (str) or -9999 (int or float) if they do
        not exist are 'nsa_version', 'nsa_nsaid', 'mngtarg1',
        'mngtarg3', 'z', 'nsa_elpetro_ba', 'nsa_elpetro_phi',
        'nsa_elpetro_th50_r', 'nsa_vdisp'.
        
        Returns:
            numpy.array: 14 arrays with: MaNGA ID, object right
            ascension, object declination, catalog ID, index of the
            entry in the catalog, catalog version, ID of object in the
            catalog, main MaNGA survey target bitmask, ancillary MaNGA
            survey target bitmask, velocity, velocity dispersion,
            ellipticity, position angle, and effective radius.
        """
        # Open the drpall file
        hdu = fits.open(self.drpall)
        # Check the plateifus are unique
        pltifu = hdu[1].data['PLATEIFU']
        if len(numpy.unique(pltifu)) != len(pltifu):
            raise ValueError('The PLATEIFU column in the DRPall file is not unique!')

        # Find the rows in the DRPall file with the correct; assumes the 
        this_pltifu = numpy.array(['{0}-{1}'.format(p,i) 
                                    for p,i in zip(self.platelist,self.ifudesignlist)])
        rows = numpy.array([numpy.where(pltifu == pi)[0][0] if pi in pltifu else -1 \
                                for pi in this_pltifu])
        if numpy.any(rows < 0):
            raise ValueError('Could not find the following in the DRPall file: '
                             '{0}'.format(this_pltifu[rows < 0]))
        indx = rows > -1
        rows = rows[indx]

        # Initialize the data
        n_drp = len(self.platelist)
        mangaid = numpy.full(n_drp, 'NULL', dtype=object)
        objra = numpy.full(n_drp, -9999.0, dtype=float)
        objdec = numpy.full(n_drp, -9999.0, dtype=float)
        catid = numpy.full(n_drp, -9999, dtype=int)
        catindx = numpy.full(n_drp, -9999, dtype=int)
        trg_version = numpy.full(n_drp, 'NULL', dtype=object)
        trg_id = numpy.full(n_drp, -9999, dtype=int)
        manga_trg1 = numpy.full(n_drp, -9999, dtype=int)
        manga_trg3 = numpy.full(n_drp, -9999, dtype=int)
        vel = numpy.full(n_drp, -9999.0, dtype=float)
        veldisp = numpy.full(n_drp, -9999.0, dtype=float)
        ell = numpy.full(n_drp, -9999.0, dtype=float)
        pa = numpy.full(n_drp, -9999.0, dtype=float)
        Reff = numpy.full(n_drp, -9999.0, dtype=float)

        # Get the data
        mangaid[indx] = hdu[1].data['mangaid'][rows]
        objra[indx] = hdu[1].data['objra'][rows]
        objdec[indx] = hdu[1].data['objdec'][rows]
        for i in numpy.where(indx)[0]:
            catid[i], catindx[i] = map(lambda x: x.astype(int).tolist(),
                                        numpy.array(mangaid[i].split('-')))
        trg_version[indx] = hdu[1].data['nsa_version'][rows]
        trg_id[indx] = hdu[1].data['nsa_nsaid'][rows]
        manga_trg1[indx] = hdu[1].data['mngtarg1'][rows]
        manga_trg3[indx] = hdu[1].data['mngtarg3'][rows]
        vel[indx] = hdu[1].data['z'][rows] * astropy.constants.c.to('km/s').value
        # There is no vdisp
        ell[indx] = 1 - hdu[1].data['nsa_elpetro_ba'][rows]
        pa[indx] = hdu[1].data['nsa_elpetro_phi'][rows]
        Reff[indx] = hdu[1].data['nsa_elpetro_th50_r'][rows]

        # Done
        hdu.close()
        return mangaid.astype(str), objra, objdec, catid, catindx, trg_version.astype(str), \
                    trg_id, manga_trg1, manga_trg3, vel, veldisp, ell, pa, Reff

    def _all_data_exists(self):
        """
        Determines if the data for all the plates/ifudesigns selected in
        the current compilation is already present in the current
        drpcomplete fits file.

        Returns:
            :obj:`bool`: Flag that all data has already been collated
            for the requested plate/ifudesign list.
        """
        if not self._confirm_access():
            return False
        for p,i in zip(self.platelist, self.ifudesignlist):
            if numpy.sum((self['PLATE'] == p) & (self['IFUDESIGN'] == i)) == 0:
                return False
        return True

    def _read(self):
        """Read the data in the existing file at :func:`file_path`."""
        if self.hdu is not None:
            self.hdu.close()
            self.hdu = None
        self.hdu = fits.open(self.file_path())
        self.nobs = self.hdu['DRPC'].header['NAXIS2']
        print('Read data: {0} rows'.format(self.nobs))

    def _confirm_access(self, reread=False):
        """
        Check the drpcomplete fits file at :func:`file_path` is
        accessible, and read the data if not yet read.

        Args:
            reread (:obj:`bool`, optional):
                Force the file to be re-read
        
        Returns:
            :obj:`bool`: Flag that :attr:`hdu` is read and valid.
        """
        if not os.path.exists(self.file_path()):
            return False
        if self.hdu is None or reread:
            self._read()
        return True

    def _find_completed_reductions(self, mindesign=19, combinatorics=False, on_disk=False):
        """
        Search the DRP path for reduced CUBE files.

        Function allows for one or both of :attr:`platelist` and
        :attr:`ifudesignlist` to be None.  The behavior is:
        
            - If both are None, all available CUBE files are used to
              create :attr:`platelist` and :attr:`ifudesignlist`.

            - If one is None, all available CUBE files within the
              provided list of the one that is not None are used to
              create :attr:`platelist` and :attr:`ifudesignlist`.  E.g.,
              if :attr:`platelist` =[7443] and :attr:`ifudesignlist` is
              None, all CUBE files with plate=7443 are chosen.

            - If both are not None and they have different lengths, or
              the same length and combinatorics is True, all available
              CUBE files within the constraints of both lists are
              selected. E.g., if :attr:`platelist` =[7443,7495] and
              :attr:`ifudesignlist` =[12704], the CUBE files with
              (plate,ifudesign)=[(7443,12704),(7459,12704)] are chosen.

            - If both are not None and they have the same length and
              combinatorics is False, all available CUBE files in the
              matched lists are chosen.  E.g. if :attr:`platelist`
              =[7443,7495] and :attr:`ifudesignlist` =[12704,12703], the
              CUBE files with
              (plate,ifudesign)=[(7443,12704),(7459,12703) are chosen

        Args:
            mindesign (:obj:`int`, optional):
                Minimum bundle design to consider.  For example, to
                ignore all the 7-fiber bundles, set `mindesign=19`
                (default) to only select the bundles with 19 or more
                fibers.
            combinatorics (:obj:`bool`, optional):
                Create :attr:`platelist` and :attr:`ifudesignlist` by
                determining all possible combinations of the input
                values. See above.
            on_disk (:obj:`bool`, optional):
                When searching for available files to analyze, search
                the DRP directory path instead of using the data in the
                DRPall file.

        Returns:
            list: Two lists with the available plates and ifudesigns for
            which to collect data.
        """
        # Are the plate and IFU lists matched?
        matchedlist = (self.platelist is not None and self.ifudesignlist is not None \
                       and (len(self.platelist) == len(self.ifudesignlist) and not combinatorics))

        # Get the list of files
        if matchedlist:
            # Lists already matched, just construct the file names
            files = [os.path.join(*MaNGADataCube.default_paths(p, i, drpver=self.drpver,
                                                               redux_path=self.redux_path)) \
                        for p,i in zip(self.platelist, self.ifudesignlist)]
        elif on_disk:
            # Find the DRP LOGCUBE files on disk
            print('Searching for completed DRP CUBE files...', end='\r')
            files = glob.glob(os.path.join(self.redux_path, '*', 'stack', '*-LOGCUBE.fits.gz'))
            print('Searching for completed DRP CUBE files...DONE.')
        else:
            # Use the DRPall file
            drpall_hdu = fits.open(self.drpall)
            pltifu = drpall_hdu[1].data['PLATEIFU']
            if len(numpy.unique(pltifu)) != len(pltifu):
                raise ValueError('The PLATEIFU column in the DRPall file is not unique!')
            files = [os.path.join(*MaNGADataCube.default_paths(int(p), int(i), drpver=self.drpver,
                                                               redux_path=self.redux_path)) \
                        for p,i in zip(drpall_hdu[1].data['plate'], drpall_hdu[1].data['ifudsgn'])]

        # Only use those files that exist
        for i in range(len(files)):
            if not os.path.isfile(files[i]):
                warnings.warn('No such file: {0}'.format(files[i]))
                del files[i]

        # Get the list of plate and ifus
        pltifu = numpy.array(list(map(lambda x: os.path.split(x)[-1].split('-')[1:3],
                                      files))).astype(int)

        # Ignore any below the minimum ifusize
        pltifu = pltifu[numpy.invert(pltifu[:,1]//100 < mindesign),:]
        if matchedlist:
            return pltifu[:,0], pltifu[:,1]

        # Select those that match to the provided list
        indx = numpy.ones(len(pltifu), dtype=bool) if self.platelist is None \
                    else numpy.array([p in self.platelist for p in pltifu[:,0]])
        indx &= numpy.ones(len(pltifu), dtype=bool) if self.ifudesignlist is None \
                    else numpy.array([i in self.ifudesignlist for i in pltifu[:,1]])
        return pltifu[indx,0], pltifu[indx,1]
        
    def _find_modes(self):
        """
        Using the provided list of CUBE DRP files, find if the RSS mode
        is also available.

        .. warning::
            - This assumes everything in :attr:`platelist` and
              :attr:`ifudesignlist` has a 'CUBE' file.

        Currently only two modes are possible:

            - (1) only 'CUBE' file are available

            - (2) both 'CUBE' and 'RSS' files are available

        Returns:
            `numpy.ndarray`: Array of modes for each input DRP file.
        """
        print('Checking for RSS counterparts...', end='\r')
        has_rss = [os.path.isfile(os.path.join(*MaNGARSS.default_paths(p, i, drpver=self.drpver,
                                                                    redux_path=self.redux_path)))
                        for p,i in zip(self.platelist, self.ifudesignlist)]
        print('Checking for RSS counterparts...DONE.')
        modes = numpy.ones(len(has_rss), dtype=int)
        modes[has_rss] = 2
        return modes
        
    def _write_parameter_struct(self, ostream):
        """
        Write the structure of the DAP SDSS-style parameter set to a
        file.

        Args:
            ostream (_io.TextIOWrapper) : File stream for output
        
        """
        ostream.write('typedef struct {\n')
        ostream.write('    long plate;\n')
        ostream.write('    long ifudesign;\n')
        ostream.write('    char mode[4];\n')
        ostream.write('    double vel;\n')
        ostream.write('    double vdisp;\n')
        ostream.write('    double ell;\n')
        ostream.write('    double pa;\n')
        ostream.write('    double reff;\n')
        ostream.write('} DAPPAR;\n')

    def _write_parameter_list(self, ostream, index, mode):
        """
        Write a DAPPAR entry to the SDSS-style parameter file.

        Args:
            ostream (_io.TextIOWrapper) : File stream for output
            index (int) : Index of the entry in :attr:`data` to write
            mode (int) : Mode of the entry; see :func:`_find_modes`

        """
        ostream.write('DAPPAR {0:4d} {1:5d} {2:4s} {3:14.7e} {4:14.7e} {5:14.7e} {6:14.7e}'
                      ' {7:14.7e}\n'.format(self.hdu['DRPC'].data['PLATE'][index],
                                            self.hdu['DRPC'].data['IFUDESIGN'][index], mode,
                                            self.hdu['DRPC'].data['VEL'][index],
                                            self.hdu['DRPC'].data['VDISP'][index],
                                            self.hdu['DRPC'].data['ELL'][index],
                                            self.hdu['DRPC'].data['PA'][index],
                                            self.hdu['DRPC'].data['REFF'][index]))

    # ******************************************************************
    #  User functions
    # ******************************************************************
    def file_name(self):
        """Return the name of the DRP complete database."""
        return ('drpcomplete_{0}.fits'.format(self.drpver))

    def file_path(self):
        """Return the full pat to the DRP complete database."""
        return os.path.join(self.directory_path, self.file_name())

    def update(self, platelist=None, ifudesignlist=None, combinatorics=False, force=False,
               alldrp=False, use_platetargets=False, on_disk=False, quiet=False):
        """
        Update the DRP complete file.
        
        If `platelist` and/or `ifudesignlist` are provided, the existing
        :attr:`platelist` and :attr:`ifudesignlist` attributes are
        replaced.

        If `platelist` and `ifudesignlist` do not have the same length
        or `combinatorics` is True, the lists are expanded to include
        all combinations of their elements.

        If `platelist` and `ifudesignlist` are None or `alldrp` is True,
        all available plates/ifudesigns are collected from within the
        DRP directory structure.

        If the result of :func:`file_path` does not exist, it is
        created.

        If the result of :func:`file_path` does exist, the available
        plates and ifudesigns in the file are compared against the list
        (provided or collected) to update.  If the lists differ, the
        drpcomplete fits file is re-created from scratch.  If all the
        plates and ifudesigns are available, nothing is done, unless
        `force=True`.

        Args:
            platelist (:obj:`str`, :obj:`list`, optional):
                List of plates to include in the drpcomplete fits file.
            ifudesignlist (:obj:`str`, :obj:`list`, optional):
                List of ifudesigns to include in the drpcomplete fits
                file.
            combinatorics (:obj:`bool`, optional):
                Determine all combinations of the entered plates and
                ifudesigns.
            force (:obj:`bool`, optional):
                Overwrite any existing drpcomplete fits file with a new
                one built from scratch.
            alldrp (:obj:`bool`, optional):
                Find the full list of available DRP files.
            use_platetargets (:obj:`bool`, optional):
                Generate the data using the platetargets files instead
                of the DRPall file.
            on_disk (:obj:`bool`, optional):
                When searching for available files to analyze, search
                the DRP directory path instead of using the data in the
                DRPall file.
            quiet (:obj:`bool`, optional):
                Suppress terminal output

        Raises:
            ValueError:
                Raised if drpcomplete fits file was opened in read-only
                mode.

        """
        if self.readonly:
            raise ValueError('drpcomplete fits file was opened as read-only!')

        if platelist is not None:
            self.platelist = arginp_to_list(platelist, evaluate=True)
        if ifudesignlist is not None:
            self.ifudesignlist = arginp_to_list(ifudesignlist, evaluate=True)

        if alldrp:
            self.platelist = None
            self.ifudesignlist = None

        # This *only* searches for CUBE files
        self.platelist, self.ifudesignlist \
                = self._find_completed_reductions(combinatorics=combinatorics, on_disk=on_disk)

        # Check if these plate-ifus are already in the table.
        if not force and self._all_data_exists():
            print('{0} up to date.'.format(self.file_path()))
            return
        else:
            print('Updating {0}.'.format(self.file_path()))

        # Past this point, the drpcomplete fits file will be ovewritten
        # if it already exists.  Only DRP files defined by 
        # self.platelist and self.ifudesignlist will be included; i.e.,
        # this does *not* append to the file but completely overwrites
        # it.

        # Check for the RSS files
        modes = self._find_modes()

        # Notify
        print('Number of DRP files for DRPComplete file: {0}'.format(len(self.platelist)))

        # Get the data
        mangaid, objra, objdec, catid, catindx, trg_version, trg_id, manga_trg1, manga_trg3, vel, \
                veldisp, ell, pa, Reff = (self._match_platetargets(quiet=quiet)
                                            if use_platetargets else self._match_drpall())

        # Apply any corrections to the redshifts using the redshift fix file
        redshift_fix_data = self._read_redshift_fix()
        if redshift_fix_data is not None:
            fix_pltifu = numpy.array(['{0}-{1}'.format(p,i) 
                                    for p,i in zip(redshift_fix_data['DAPZCORR']['plate'],
                                                   redshift_fix_data['DAPZCORR']['ifudesign'])])
            this_pltifu = numpy.array(['{0}-{1}'.format(p,i) 
                                       for p,i in zip(self.platelist,self.ifudesignlist)])
            rows = numpy.array([numpy.where(fix_pltifu == pi)[0][0] if pi in fix_pltifu else -1 \
                                for pi in this_pltifu])
            indx = rows > -1
            rows = rows[indx]
            vel[indx] = redshift_fix_data['DAPZCORR']['z'][rows] \
                            * astropy.constants.c.to('km/s').value

        photometry_fix_data = self._read_photometry_fix()
        if photometry_fix_data is not None:
            fix_pltifu = numpy.array(['{0}-{1}'.format(p,i) 
                                    for p,i in zip(photometry_fix_data['DAPPHOTCORR']['plate'],
                                                   photometry_fix_data['DAPPHOTCORR']['ifudesign'])])
            this_pltifu = numpy.array(['{0}-{1}'.format(p,i) 
                                       for p,i in zip(self.platelist,self.ifudesignlist)])
            rows = numpy.array([numpy.where(fix_pltifu == pi)[0][0] if pi in fix_pltifu else -1 \
                                for pi in this_pltifu])
            indx = rows > -1
            rows = rows[indx]
            ell[indx] = photometry_fix_data['DAPPHOTCORR']['ell'][rows]
            pa[indx] = photometry_fix_data['DAPPHOTCORR']['pa'][rows]
            Reff[indx] = photometry_fix_data['DAPPHOTCORR']['reff'][rows]

        # Write the data to disk
        self.write(self.platelist, self.ifudesignlist, modes, mangaid, objra, objdec, catid,
                   catindx, trg_version, trg_id, manga_trg1, manga_trg3, vel, veldisp, ell, pa,
                   Reff)

    def write(self, platelist, ifudesignlist, modes, mangaid, objra, objdec, catid, catindx,
              trg_version, trg_id, manga_trg1, manga_trg3, vel, veldisp, ell, pa, Reff, drpver=None,
              redux_path=None, dapver=None, analysis_path=None, overwrite=True):
        r"""
        Write the drpcomplete fits binary table.

        Header keywords are:
            - VERSDRP: DRP version
            - RDXPTH: DRP reduction path
            - VERSDAP: DAP version
            - SISPTH: DAP analysis path
            - PLTTRG *N*: plateTargets file *N*
            - CATID *N*: ID Number of target catalog *N*
            - AUTHOR: Set to 'K.B. Westfall <kbwestfall@gmail.com>'

        Binary table columns:
            1. PLATE (1J): Plate number
            2. IFUDESIGN (1J): IFU design
            3. MODES (1B): Modes available 
                - MODES=1: only 'CUBE' file is available
                - MODES=2: both 'CUBE' and 'RSS' files are available

            4. MANGAID (*n* A): MaNGA ID (same as 'CATID-CATINDX')
            5. OBJRA (1D): Object right ascension
            6. OBJDEC (1D): Object declination
            7. CATID (1J): Catalog ID used for target selection
            8. CATINDX (1J): Index in catalog with target data
            9. TRG_VERSION (*n* A): Version of the catalog (e.g., NSA)
                used in targetting (def='NULL' if not available)
            10. TRG_ID (1J): Target ID in catalog (def=-9999 if not
                available)
            11. MANGA_TARGET1: Main MaNGA survey target bitmask
            12. MANGA_TARGET3: Ancillary MaNGA survey target bitmask
            13. VEL (1D): Systemic velocity from catalog (def=-9999.0 if
                not available)
            14. VDISP (1D): Velocity dispersion from catalog
                (def=-9999.0 if not available)
            15. ELL (1D): Ellipticity (1-b/a) from catalog (def=-9999.0
                if not available); Elliptical Petrosian value from NSA
            16. PA (1D): Position angle from catalog (def=-9999.0 if not
                available); Elliptical Petrosian value from NSA
            17. REFF (1D): Effective radius from catalog (def=-9999.0 if
                not available); Elliptical Petrosian value from NSA

        Args:
            platelist (list) : List of plates
            ifudesignlist (list) : List of IFU designs
            modes (numpy.array): Mode values, see above
            mangaid (numpy.array): MaNGA IDs
            objra (numpy.array): Object right ascensions
            objdec (numpy.array): Object declinations
            catid (numpy.array): Catalog ID used for target selection
            catindx (numpy.array): Index in catalog with target data
            trg_version (numpy.array): Version of the catalog (e.g.,
                NSA) used in targetting
            trg_id (numpy.array): Target ID in catalog
            manga_trg1 (numpy.array): Main MaNGA survey target bitmask
            manga_trg3 (numpy.array): Ancillary MaNGA survey target
                bitmask
            vel (numpy.array): Velocity from catalog, if available
            veldisp (numpy.array): Velocity dispersion from catalog,
                if available
            ell (numpy.array): Ellipticity (1-b/a) from catalog, if
                available
            pa (numpy.array): Position angle from catalog, if
                available
            Reff (numpy.array): Effective radius from catalog, if
                available
            drpver (str) : (**Optional**) DRP version, see above.
            redux_path (str) : (**Optional**) Path to the top level of
                directory for the DRP output, see above.
            dapver (str) : (**Optional**) DAP version, see above.
            analysis_path (str) : (**Optional**) Path to the top level
                directory for the DAP output files, see above.
            overwrite (bool): (**Optional**) Overwrite any existing file.

        Raises:
            AttributeError: Raised if drpcomplete fits file was opened
                in read-only mode.
            FileExistsError: Raised if the drpcomplete file exists and
                overwrite=False.
        """
        if self.readonly:
            raise AttributeError('drpcomplete fits file was opened as read-only!')

        out=self.file_path()
        if os.path.isfile(out) and not overwrite:
            raise FileExistsError('DRP complete file already exists: {0}'.format(out))

        if not os.path.isdir(self.directory_path):
            os.makedirs(self.directory_path)

        # Create the primary header
        nplttrg = 0 if self.platetargets is None else len(self.platetargets)

        hdr = fits.Header()
        hdr['VERSDRP'] = self.drpver if drpver is None else drpver
        hdr['RDXPTH'] = self.redux_path if redux_path is None else redux_path
        hdr['VERSDAP'] = self.dapver if dapver is None else dapver
        hdr['SISPTH'] = self.analysis_path if analysis_path is None else analysis_path
        for i in range(nplttrg):
            hdr['PLTTRG{0}'.format(i+1)] = self.platetargets[i]
            hdr['CATID{0}'.format(i+1)] = self.catid[i]
        hdr['AUTHOR'] = 'K.B. Westfall <kbwestfall@gmail.com>'

        # Create the Binary Table
        cols = []
        cols.append(fits.Column(name='PLATE', format='1J', array=numpy.array(platelist)))
        cols.append(fits.Column(name='IFUDESIGN', format='1J', array=numpy.array(ifudesignlist)))
        cols.append(fits.Column(name='MODES', format='1B', array=modes))
        cols.append(fits.Column(name='MANGAID', format='{0}A'.format(mangaid.dtype.itemsize),
                           array=mangaid))
        cols.append(fits.Column(name='OBJRA', format='1D', array=objra))
        cols.append(fits.Column(name='OBJDEC', format='1D', array=objdec))
        cols.append(fits.Column(name='CATID', format='1J', array=catid))
        cols.append(fits.Column(name='CATINDX', format='1J', array=catindx))
        cols.append(fits.Column(name='TRG_VERSION',
                                format='{0}A'.format(trg_version.dtype.itemsize),
                                array=trg_version))
        cols.append(fits.Column(name='TRG_ID', format='1J', array=trg_id))
        cols.append(fits.Column(name='MANGA_TARGET1', format='1J', array=manga_trg1))
        cols.append(fits.Column(name='MANGA_TARGET3', format='1J', array=manga_trg3))
        cols.append(fits.Column(name='VEL', format='1D', array=vel))
        cols.append(fits.Column(name='VDISP', format='1D', array=veldisp))
        cols.append(fits.Column(name='ELL', format='1D', array=ell))
        cols.append(fits.Column(name='PA', format='1D', array=pa))
        cols.append(fits.Column(name='REFF', format='1D', array=Reff))

        # Create the HDUList and write it to the fits file
        self.hdu = fits.HDUList([ fits.PrimaryHDU(header=hdr),
                                  fits.BinTableHDU.from_columns(cols, name='DRPC') ])
        print('Writing to disk: {0}'.format(out))
        self.hdu.writeto(out, overwrite=overwrite)
        self.nobs = self.hdu['DRPC'].header['NAXIS2']

#    def grab_data(self, plate=None, ifudesign=None, index=None, reread=False):
#        """
#        Return the data for a given plate-ifudesign, or index in the
#        :attr:`data`.  Even though *plate*, *ifudesign*, and *index* are
#        all optional, the arguments must contain either *plate* and
#        *ifudesign* or *index*.
#
#        Args:
#            plate (int): (**Optional**) Plate number
#            ifudesign (int): (**Optional**) IFU design
#            index (int): (**Optional**) Index of the row in :attr:`data`
#                with the data to return
#            reread (bool): (**Optional**) Force the database to be re-read
#
#        Returns:
#            list: List of the 14 elements of :attr:`data`; see
#            :func:`write`.
#
#        Raises:
#            ValueError: Raised if the row with the data is unknown
#                because either *index* is not defined or one or both of
#                *plate* and *ifudesign* is not defined.  Also raised if
#                *index* does not exist in the data array.
#        """
#        if (plate is None or ifudesign is None) and index is None:
#            raise ValueError('Must provide plate and ifudesign or row index!')
#
#        if index is None:
#            index = self.entry_index(plate, ifudesign, reread=reread)
#        else:
#            self._confirm_access(reread=reread)
#            if index >= self.nobs:
#                raise ValueError('Selected row index does not exist')
#
#        # TODO: Can't I just select index, i.e.,
#        # self.hdu[1].data[index]?
#
#        return [ self.hdu['DRPC'].data['PLATE'][index], self.hdu['DRPC'].data['IFUDESIGN'][index],
#                 self.hdu['DRPC'].data['MODES'][index], self.hdu['DRPC'].data['MANGAID'][index],
#                 self.hdu['DRPC'].data['OBJRA'][index], self.hdu['DRPC'].data['OBJDEC'][index],
#                 self.hdu['DRPC'].data['CATID'][index], self.hdu['DRPC'].data['CATINDX'][index],
#                 self.hdu['DRPC'].data['TRG_VERSION'][index], self.hdu['DRPC'].data['TRG_ID'][indx],
#                 self.hdu['DRPC'].data['NSA_V100_ID'][index], self.hdu['DRPC'].data['VEL'][index],
#                 self.hdu['DRPC'].data['VDISP'][index], self.hdu['DRPC'].data['ELL'][index],
#                 self.hdu['DRPC'].data['PA'][index], self.hdu['DRPC'].data['REFF'][index] ]

    def write_par(self, ofile, mode, plate=None, ifudesign=None, index=None, reread=False,
                  overwrite=True):
        """
        Write the SDSS-style parameter (Yanny) file for use with the
        MaNGA DAP.

        Args: 
            ofile (str): Output file name
            mode (str): Mode of the DRP file to analyze; must be either
                'RSS' or 'CUBE'
            plate (int): (**Optional**) Plate number
            ifudesign (int): (**Optional**) IFU design
            index (int): (**Optional**) Index of the row in :attr:`data`
                with the data to return
            reread (bool): (**Optional**) Force the database to be re-read
            overwrite (bool): (**Optional**) Overwrite any existing parameter
                file

        Raises:
            IOError: Raised if the parameter file already exists and
                overwrite is False.
            ValueError: Raised if

                - the row with the data is unknown because either
                  *index* is not defined or one or both of *plate* and
                  *ifudesign* is not defined.
                - *index* does not exist in the data array.
                - the 'RSS' mode is selected but unavailable

        """
        if not self._confirm_access(reread=reread):
            raise IOError('Could not access database!')

        if os.path.exists(ofile) and not overwrite:
            raise IOError('Parameter file already exists.  Set overwrite=True to overwrite.')

        if (plate is None or ifudesign is None) and index is None:
            raise ValueError('Must provide plate and ifudesign or row index!')

        if mode != 'CUBE' and mode != 'RSS':
            raise ValueError('Mode must be either CUBE or RSS.')

        if index is None:
            index = self.entry_index(plate, ifudesign, reread=reread)
        else:
            if index >= self.nobs:
                raise ValueError('Selected row index does not exist')

        if mode == 'RSS' and self.hdu['DRPC'].data['MODES'][index] != 2:
            raise ValueError('RSS mode not available for plate={0},ifudesign={1} !'.format(
                                                        self.hdu['DRPC'].data['PLATE'][index],
                                                        self.hdu['DRPC'].data['IFUDESIGN'][index]))

        # Write the SDSS parameter file
        ostream = open(ofile, 'w')
        ostream.write('\n')
        self._write_parameter_struct(ostream)
        ostream.write('\n')
        ostream.write('\n')
        self._write_parameter_list(ostream, index, mode)
        ostream.write('\n')
        ostream.close()

    def write_config(self, ofile, plate=None, ifudesign=None, index=None, sres_ext=None,
                     sres_fill=None, covar_ext=None, reread=False, overwrite=True):
        """
        Write a config file with the data used to instantiate a
        :class:`mangadap.datacube.manga.MaNGADataCube` datacube for
        analysis.

        Args: 
            ofile (:obj:`str`):
                Output file name.
            plate (:obj:`int`, optional):
                Plate number.
            ifudesign (:obj:`int`, optional):
                IFU design.
            index (:obj:`int`, optional):
                Index of the row in :attr:`data` with the data to
                return.
            sres_ext (:obj:`str`, optional):
                The extension to use when constructing the spectral
                resolution vectors for the MaNGA datacubes. See
                :func:`mangadap.datacube.manga.MaNGADataCube.spectral_resolution`.
            sres_fill (:obj:`bool`, optional):
                Fill masked spectral-resolution data by simple linear
                interpolation.
            covar_ext (:obj:`str`, optional):
                Extension in the MaNGA DRP CUBE file to use as the
                single spatial correlation matrix for all wavelength
                channels.
            reread (:obj:`bool`, optional):
                Force the database to be re-read
            overwrite (:obj:`bool`, optional):
                Overwrite any existing parameter file

        Raises:
            IOError:
                Raised if the parameter file already exists and
                overwrite is False.
            ValueError:
                Raised if
                    - the row with the data is unknown because either
                      ``index`` is not defined or one or both of
                      ``plate`` and ``ifudesign`` is not defined.
                    - ``index`` does not exist in the data array.

        """
        if not self._confirm_access(reread=reread):
            raise IOError('Could not access database!')

        if (plate is None or ifudesign is None) and index is None:
            raise ValueError('Must provide plate and ifudesign or row index!')

        if index is None:
            index = self.entry_index(plate, ifudesign, reread=reread)
        elif index >= self.nobs:
            raise ValueError('Selected row index does not exist')

        MaNGADataCube.write_config(ofile, self['PLATE'][index], self['IFUDESIGN'][index], log=True,
                                   z=self['VEL'][index]/astropy.constants.c.to('km/s').value,
                                   vdisp=self['VDISP'][index], ell=self['ELL'][index],
                                   pa=self['PA'][index], reff=self['REFF'][index],
                                   sres_ext=sres_ext, sres_fill=sres_fill, covar_ext=covar_ext,
                                   drpver=self.drpver, redux_path=self.redux_path,
                                   overwrite=overwrite)

    def entry_index(self, plate, ifudesign, reread=False):
        """
        Find the index of the row with the parameter data for the
        specified plate and ifudesign.

        .. warning::
            - This is very inefficient if you're looking for multiple
              entries...

        Args:
            plate (:obj:`int`):
                Plate number
            ifudesign (:obj:`int`):
                IFU design
            reread (:obj:`bool`, optional):
                Force the database to be re-read

        Returns:
            int: Index of the row in :attr:`data` with the data for the
            given *plate* and *ifudesign*

        Raises:
            ValueError:
                Raised if the given `plate` and `ifudesign` were not
                found.
        """
        if not self._confirm_access(reread=reread):
            raise IOError('Could not access database!')
        indx = (self['PLATE'] == plate) & (self['IFUDESIGN'] == ifudesign)
        if numpy.sum(indx) == 0:
            raise ValueError('Could not find plate={0},ifudesign={1} in drpcomplete file!'.format(
                                plate, ifudesign))
        return numpy.where(indx)[0][0]

    def can_analyze(self, row=None):
        """
        Determine if the DAP can analyze a plate-ifu entry in the
        database.

        The selection is currently:

            - MaNGAID != 'NULL'
            - MANGA_TARGET1 > 0 or MANGA_TARGET3 > 0
            - VEL > -500

        Args:
            row (:obj:`int`, optional):
                The specific row to test.  By default, return a boolean
                vector for all the database rows.

        Returns:
            Either a single boolean or boolean `numpy.ndarray`_
            flagging that DAP can (True) or cannot (False) analyze
            the data associated with the database entry (or entries).
        """
        if row is None:
            return (self['MANGAID'] != 'NULL') \
                    & ((self['MANGA_TARGET1'] > 0) | (self['MANGA_TARGET3'] > 0)) \
                    & (self['VEL'] > -500.0)
        return self['MANGAID'][row] != 'NULL' \
                and (self['MANGA_TARGET1'][row] > 0 or self['MANGA_TARGET3'][row] > 0) \
                and self['VEL'][row] > -500.0




from pathlib import Path

from IPython import embed

from mangadap.config.manga import MaNGAConfig
from mangadap.tests.util import data_test_file, remote_data_file, requires_remote


def test_init():
    cfg = MaNGAConfig(7815, 3702)
    assert isinstance(cfg.directory_path, Path), 'Directory should be a Path instance'
    assert cfg.plate == 7815, 'Plate changed'
    assert cfg.log, 'Log binning should be true'
    assert cfg.mode == 'CUBE', 'Should be in CUBE mode'
    assert cfg.file_name == 'manga-7815-3702-LOGCUBE.fits.gz'


def test_copy():
    cfg = MaNGAConfig(7815, 3702)
    _cfg = cfg.copy()
    assert cfg.file_name == _cfg.file_name, 'Bad name copy'
    _cfg.mode = 'RSS'
    assert cfg.file_name != _cfg.file_name, 'Bad name change'


def test_from_config():
    cfg = MaNGAConfig.from_config(data_test_file("datacube.ini"))
    assert isinstance(cfg.directory_path, Path), 'Directory should be a Path instance'
    # TODO: This passes locally, but will fail the tox CI tests.  Determine how
    # to perform this in tox?
    #assert cfg.directory_path == Path(remote_data_file()).resolve(), 'Directory path changed'
    assert cfg.plate == 7815, 'Plate changed'
    assert cfg.log, 'Log binning should be true'
    assert cfg.mode == 'CUBE', 'Should be in CUBE mode'
    assert cfg.file_name == 'manga-7815-3702-LOGCUBE.fits.gz'


@requires_remote
def test_from_file():
    file = remote_data_file('manga-7815-3702-LOGCUBE.fits.gz')
    cfg = MaNGAConfig.from_file(file)

    assert isinstance(cfg.directory_path, Path), 'Directory should be a Path instance'
    assert cfg.directory_path == Path(remote_data_file()).resolve(), 'Directory path changed'
    assert cfg.plate == 7815, 'Plate changed'
    assert cfg.log, 'Log binning should be true'
    assert cfg.mode == 'CUBE', 'Should be in CUBE mode'
    assert cfg.file_name == 'manga-7815-3702-LOGCUBE.fits.gz'





from IPython import embed

import numpy
from scipy import interpolate
from astropy.io import fits

from mangadap.util import sampling
from mangadap.proc.bandpassfilter import passband_integral
from mangadap.tests.util import data_test_file


def test_resample():
    delt = 100
    x = numpy.arange(20)*0.5 + delt
    y = numpy.linspace(0.1, 2.0, 20)
    y = y*y - y*y*y + 0.2*y + 3 + 0.2 * numpy.square(numpy.sin(50*y))
    newRange = numpy.array([2,8]) + delt
    r = sampling.Resample(y, xRange=[x[0],x[-1]], newRange=newRange, newpix=40, newLog=True,
                          step=False)
    assert numpy.isclose(r.outx[0], newRange[0]), 'Incorrect starting value'
    assert numpy.mean(interpolate.interp1d(x,y)(r.outx) - r.outy) < 1e-3, \
                'Bad interpolation difference'

    _r = sampling.Resample(y, xRange=[x[0],x[-1]], newBorders=r.outborders, step=False)
    assert numpy.array_equal(r.outx, _r.outx), 'Coordinates differ'
    assert numpy.array_equal(r.outborders, _r.outborders), 'Coordinates borders differ'
    assert numpy.array_equal(r.outy, _r.outy), 'Resampling differs'

    _r = sampling.Resample(y, xRange=[x[0],x[-1]], newx=r.outx, newLog=True, step=False)
    assert numpy.array_equal(r.outx, _r.outx), 'Coordinates differ'
    assert numpy.allclose(r.outborders, _r.outborders, rtol=0., atol=1e-12), \
                'Coordinates borders differ'
    assert numpy.allclose(r.outy, _r.outy, rtol=0., atol=1e-12), \
                'Coordinates borders differ'
                
    rc = sampling.Resample(y, xRange=[x[0],x[-1]], newRange=newRange, newpix=40, newLog=True,
                           conserve=True, step=False)
    assert numpy.isclose(numpy.mean(r.outy/rc.outy), 3.250907334), \
                'Flux conservation test failed'


def test_resample_covar():
    delt = 100
    x = numpy.arange(20)*0.5 + delt
    y = numpy.linspace(0.1, 2.0, 20)
    y = y*y - y*y*y + 0.2*y + 3 + 0.2 * numpy.square(numpy.sin(50*y))
    newRange = numpy.array([2,8]) + delt
    e = numpy.full_like(y, 2.)
    r = sampling.Resample(y, e=e, xRange=[x[0],x[-1]], newRange=newRange, newpix=40, newLog=True)

    # Test covar without errors
    _r = sampling.Resample(y, xRange=[x[0],x[-1]], newRange=newRange, newpix=40, newLog=True,
                           covar=True)
    assert numpy.array_equal(r.outy, _r.outy), 'Bad resampling with covariance'

    # Test multiple vectors
    _r = sampling.Resample(numpy.row_stack((y,y)), xRange=[x[0],x[-1]], newRange=newRange,
                           newpix=40, newLog=True, covar=True)
    assert numpy.allclose(r.outy, _r.outy[0], rtol=0.0, atol=1e-15), \
            'Bad resampling with covariance'

    # Test covar with errors
    _r = sampling.Resample(y, xRange=[x[0],x[-1]], e=e, newRange=newRange, newpix=40, newLog=True,
                           covar=True)
    assert numpy.array_equal(r.outy, _r.outy), 'Bad resampling with covariance'
    assert numpy.array_equal(r.oute, _r.oute), 'Bad resampling errors with covariance'

    # Test multiple vectors with errors
    _r = sampling.Resample(numpy.row_stack((y,y)), e=numpy.row_stack((e,e)), xRange=[x[0],x[-1]],
                           newRange=newRange, newpix=40, newLog=True, covar=True)
    assert numpy.allclose(r.outy, _r.outy[0], rtol=0.0, atol=1e-15), \
            'Bad resampling with covariance'
    assert numpy.allclose(r.oute, _r.oute[0], rtol=0.0, atol=1e-15), \
            'Bad resampling errors with covariance'

    # Test with conserve
    r = sampling.Resample(y, e=e, xRange=[x[0],x[-1]], newRange=newRange, newpix=40, newLog=True,
                          conserve=True)
    _r = sampling.Resample(y, xRange=[x[0],x[-1]], e=e, newRange=newRange, newpix=40, newLog=True,
                           covar=True, conserve=True)
    assert numpy.allclose(r.outy, _r.outy, rtol=0.0, atol=1e-15), \
            'Bad resampling with covariance'
    assert numpy.allclose(r.oute, _r.oute, rtol=0.0, atol=1e-15), \
            'Bad resampling errors with covariance'

    _r = sampling.Resample(numpy.row_stack((y,y)), e=numpy.row_stack((e,e)), xRange=[x[0],x[-1]],
                           newRange=newRange, newpix=40, newLog=True, covar=True, conserve=True)
    assert numpy.allclose(r.outy, _r.outy[0], rtol=0.0, atol=1e-15), \
            'Bad resampling with covariance'
    assert numpy.allclose(r.oute, _r.oute[0], rtol=0.0, atol=1e-15), \
            'Bad resampling errors with covariance'


def test_against_brute_force():
    specfile = data_test_file('MaNGA_test_spectra.fits.gz')
    hdu = fits.open(specfile)

    # Just test on the first spectrum
    old_wave = hdu['WAVE'].data
    old_flux = numpy.ma.MaskedArray(hdu['FLUX'].data[0,:], mask=hdu['MASK'].data[0,:] > 0)
    old_flux[(old_wave > 5570) & (old_wave < 5586)] = numpy.ma.masked
    old_ferr = numpy.ma.power(hdu['IVAR'].data[0,:], -0.5)
    z = hdu['Z'].data[0]

    # Do the brute force calculation
    borders = sampling.grid_borders(numpy.array([old_wave[0],old_wave[-1]]), old_wave.size,
                                    log=True)[0]
    _p = numpy.repeat(borders, 2)[1:-1].reshape(-1,2)
    new_flux_brute = passband_integral(old_wave/(1+z), old_flux, passband=_p, log=True)
    new_flux_brute /= (_p[:,1]-_p[:,0])

    # Use resample
    r = sampling.Resample(old_flux, e=old_ferr, x=old_wave/(1+z),
                          newRange=[old_wave[0], old_wave[-1]], inLog=True, newLog=True)

    # The two shoule be the same
    assert numpy.isclose(numpy.mean(numpy.absolute(new_flux_brute-r.outy)), 0.0), \
                'Resampling and brute force calculation should be identical'

def test_against_brute_force_twod():
    specfile = data_test_file('MaNGA_test_spectra.fits.gz')
    hdu = fits.open(specfile)

    # Just test on the first spectrum
    old_wave = hdu['WAVE'].data
    old_flux = numpy.ma.MaskedArray(hdu['FLUX'].data[:2,:], mask=hdu['MASK'].data[:2,:] > 0)
    old_flux[:,(old_wave > 5570) & (old_wave < 5586)] = numpy.ma.masked
    old_ferr = numpy.ma.power(hdu['IVAR'].data[:2,:], -0.5)
    z = hdu['Z'].data[0]    # Only use one redshift so that the input coordinates are the same

    # Do the brute force calculation
    borders = sampling.grid_borders(numpy.array([old_wave[0],old_wave[-1]]), old_wave.size,
                                    log=True)[0]
    _p = numpy.repeat(borders, 2)[1:-1].reshape(-1,2)
    new_flux_brute = numpy.expand_dims(passband_integral(old_wave/(1+z), old_flux[0,:],
                                                         passband=_p, log=True), axis=0)
    new_flux_brute = numpy.append(new_flux_brute, numpy.expand_dims(
                                        passband_integral(old_wave/(1+z), old_flux[1,:],
                                                          passband=_p, log=True), axis=0), axis=0) 
    new_flux_brute /= (_p[:,1]-_p[:,0])[None,:]

    # Use resample
    r = sampling.Resample(old_flux, e=old_ferr, x=old_wave/(1+z),
                          newRange=[old_wave[0], old_wave[-1]], inLog=True, newLog=True)

    # The two shoule be the same
    assert numpy.isclose(numpy.mean(numpy.absolute(new_flux_brute-r.outy)), 0.0), \
                'Resampling and brute force calculation should be identical'


def test_grid_npix():
    for s in numpy.random.randint(2, high=1000, size=100):
        x = numpy.linspace(-1.,1.,s)
        nx, rng = sampling.grid_npix(rng=[x[0],x[-1]], dx=x[1]-x[0])
        assert nx == s, 'Bad number of linear grid points'
        assert numpy.allclose(rng, [-1., 1.]), 'Bad linear grid range'

        x = numpy.logspace(-1.,1.,s)
        nx, rng = sampling.grid_npix(rng=[x[0],x[-1]], dx=numpy.log10(x[1])-numpy.log10(x[0]),
                                     log=True)
        assert nx == s, 'Bad number of logarithmic grid points'
        assert numpy.allclose(rng, numpy.power(10., [-1., 1.])), 'Bad logarithmic grid range'


def test_grid_borders():
    x = numpy.linspace(-1,1,100)
    dx = x[1]-x[0]
    nx, rng = sampling.grid_npix(rng=[x[0],x[-1]], dx=dx)
    borders, _dx = sampling.grid_borders(rng, nx)
    assert dx == _dx, 'Bad linear step size'
    assert numpy.allclose(borders, numpy.linspace(-1-dx/2, 1+dx/2, 101)), 'Bad linear borders'


    x = numpy.logspace(-1,1,100)
    dx = numpy.log10(x[1])-numpy.log10(x[0])
    nx, rng = sampling.grid_npix(rng=[x[0],x[-1]], dx=dx, log=True)
    borders, _dx = sampling.grid_borders(rng, nx, log=True)
    assert dx == _dx, 'Bad logarithmic step size'
    assert numpy.allclose(borders, numpy.logspace(-1-dx/2, 1+dx/2, 101)), 'Bad geometric borders'


def test_grid_centers():
    x = numpy.linspace(-1,1,100)
    dx = x[1]-x[0]
    nx, rng = sampling.grid_npix(rng=[x[0],x[-1]], dx=dx)
    centers, _dx = sampling.grid_centers(rng, nx)
    assert dx == _dx, 'Bad linear step size'
    assert numpy.allclose(centers, x), 'Bad linear borders'

    x = numpy.logspace(-1,1,100)
    dx = numpy.log10(x[1])-numpy.log10(x[0])
    nx, rng = sampling.grid_npix(rng=[x[0],x[-1]], dx=dx, log=True)
    centers, _dx = sampling.grid_centers(rng, nx, log=True)
    assert dx == _dx, 'Bad logarithmic step size'
    assert numpy.allclose(centers, x), 'Bad geometric borders'

if __name__ == '__main__':
    test_resample_covar()





from IPython import embed

import numpy
from mangadap.util.filter import BoxcarFilter

def test_boxcar():

    rng = numpy.random.default_rng(1001)
    wave = numpy.arange(3650.0, 10501.0, 0.5)
    x = (wave - numpy.mean(wave))/(wave[-1]-wave[0])
    sig = 1/(0.5 + 0.5*(x+0.5)**2)
    flux = 3 + 0.1*x + 0.5*x**2 + 3*(x+1)**4
    dflux = rng.normal(scale=sig)
    indx = rng.integers(x.size, size=100)
    dflux[indx] += rng.normal(scale=10*sig)[indx]

    indx = ((wave > 4567) & (wave < 4587)) \
                | ((wave > 5567) & (wave < 5587)) \
                | ((wave > 9067) & (wave < 9087))
    sim = numpy.ma.MaskedArray(flux + dflux, mask=indx)

    bf = BoxcarFilter(100, lo=3., hi=3., niter=None, y=sim, local_sigma=True)
    assert numpy.sum(bf.output_mask[0]) == 212, 'Number of rejections changed.'

#    from matplotlib import pyplot
#    pyplot.plot(wave, sim, zorder=1)
#    pyplot.plot(wave, bf.smoothed_y[0], zorder=2)
#    pyplot.scatter(wave[bf.output_mask[0]], sim[bf.output_mask[0]], marker='.', color='C3',
#                   s=100, lw=0, zorder=3)
#    pyplot.xlabel('x')
#    pyplot.ylabel('y')
#    pyplot.show()





import os
import time
import numpy

import argparse

from scipy import interpolate

from mangadap.config import defaults
from mangadap.proc.templatelibrary import TemplateLibrary
from mangadap.util.sampling import spectral_coordinate_step

from mangadap.scripts import scriptbase


class TemplateFluxNorm(scriptbase.ScriptBase):

    @classmethod
    def get_parser(cls, width=None):

        parser = super().get_parser(description='Compute flux normalizations for DAP template set',
                                    width=width)
        parser.add_argument('key', type=str, help='DAP template-library key')
        parser.add_argument('ofile', type=str, help='Output file name')
        parser.add_argument('-v', '--velscale_ratio', default=4, type=int,
                            help='velocity scale ratio between the template pixel and '
                                 'the nominal pixel width')
        parser.add_argument('-s', '--step', default=1e-4, type=float,
                            help='linear or log-linear spectral pixel width')
        parser.add_argument('-l', '--linear', action='store_true', default=False,
                            help='pixel sampling should be linear (default is log)')
        return parser

    @staticmethod
    def main(args):

        # Get the g and r filter curves
        filter_root = os.path.join(defaults.dap_data_root(), 'filter_response')

        f = os.path.join(filter_root, 'gunn_2001_g_response.db')
        if not os.path.isfile(f):
            raise FileNotFoundError('No file: {0}'.format(f))
        db = numpy.genfromtxt(os.path.join(f))
        g = interpolate.interp1d(db[:,0], db[:,1], fill_value=0.0, bounds_error=False)

        f = os.path.join(filter_root, 'gunn_2001_r_response.db')
        if not os.path.isfile(f):
            raise FileNotFoundError('No file: {0}'.format(f))
        db = numpy.genfromtxt(os.path.join(f))
        r = interpolate.interp1d(db[:,0], db[:,1], fill_value=0.0, bounds_error=False)

        # Build the template library
        tpl = TemplateLibrary(args.key, velscale_ratio=args.velscale_ratio,
                              spectral_step=args.step, log=not args.linear, hardcopy=False)

        # Get the sampling
        nwave = len(tpl['WAVE'].data)
        dwave = spectral_coordinate_step(tpl['WAVE'].data, log=not args.linear)
        if not args.linear:
            dwave *= tpl['WAVE'].data*numpy.log(10.)

        flxmid = (numpy.amax(tpl['FLUX'].data, axis=1)+numpy.amin(tpl['FLUX'].data, axis=1))/2.0
        flxrng = numpy.amax(tpl['FLUX'].data, axis=1)-numpy.amin(tpl['FLUX'].data, axis=1)

        g_tc = g(tpl['WAVE'].data)
        r_tc = r(tpl['WAVE'].data)

        g_tc_sum = numpy.sum(g_tc*dwave)
        r_tc_sum = numpy.sum(r_tc*dwave)

        gflux = numpy.sum(tpl['FLUX'].data*g_tc[None,:]*dwave[None,:], axis=1)/g_tc_sum
        rflux = numpy.sum(tpl['FLUX'].data*r_tc[None,:]*dwave[None,:], axis=1)/r_tc_sum

        numpy.savetxt(args.ofile,
                      numpy.array([ numpy.arange(tpl.ntpl), flxmid, flxrng, gflux, rflux,
                                    -2.5*numpy.log10(gflux/rflux)]).T,
                      fmt=[ '%5d', '%15.8e', '%15.8e', '%15.8e', '%15.8e', '%15.8e'],
                      header='{0:>3} {1:>15} {2:>15} {3:>15} {4:>15} {5:>15}'.format(
                                'ID', 'FLXMID', 'FLXRNG', 'MEANg', 'MEANr', '-2.5LOG(g/r)'))

